import time

from airtestProject.commons.page.odin.login import LoginPage
from airtestProject.commons.utils.gotCommon import TestGot
from airtestProject.commons.UWA import *
from airtestProject.commons.utils.logger import log
# from poco.drivers.unity3d import UnityPoco


class OdinProfile:

    def __init__(self):
        self.odin_poco_apk = "com.sanqi.odin.poco"
        self.odin_weekly_apk = "com.sanqi.odin.weekly"
        self.odin_2022weekly_apk = "com.sanqi.odin2022.weekly"
        self.odin_uwa_apk = "com.sanqi.odinUWA"
        self.odin_2022uwa_apk = "com.sanqi.odin2022.uwa"
        self.got = TestGot()
        # self.mainPage = MainPage()

    @log.wrap("跑图并上传UWA")
    def test_running_uwa(self):
        self.got.got_init(self.odin_2022uwa_apk)
        # 登录
        # self.loginPage.
        # 点击跑图
        # self.mainPage.start_running_1()
        # 开启got

        # # poco = UnityPoco()
        # self.got.got_start(poco, "default")
        # # 开始跑图
        # # self.mainPage.start_running_2()
        # # 停止got
        # self.got.got_stop(poco)
        # GOT_Test.LocalUpload(poco, account='wangmo@37.com', password='37youxi123456', projectID=1775,
        #                      timeLimitS=300)
        # 上传报告
        # self.got.got_upload(poco)
        # 杀进程
        stop_app(self.odin_poco_apk)

    @log.wrap("跑checklist")
    def test_check_list(self):
        self.got.got_init(self.odin_2022uwa_apk)
        # 登录



if __name__ == '__main__':
    odin = OdinProfile()
    odin.test_running_uwa()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""""Airtest图像识别专用."""

import os
import sys
import time
import types
from six import PY3
from copy import deepcopy

from airtestProject.airtest import aircv
from airtestProject.airtest.aircv import cv2
from airtestProject.airtest.core.helper import G, logwrap
from airtestProject.airtest.core.settings import Settings as ST  # noqa
from airtestProject.airtest.core.error import TargetNotFoundError, InvalidMatchingMethodError
from airtestProject.airtest.utils.transform import TargetPos

from airtestProject.airtest.aircv.template_matching import TemplateMatching
from airtestProject.airtest.aircv.multiscale_template_matching import MultiScaleTemplateMatching,MultiScaleTemplateMatchingPre
from airtestProject.airtest.aircv.keypoint_matching import KAZEMatching, BRISKMatching, AKAZEMatching, ORBMatching
from airtestProject.airtest.aircv.keypoint_matching_contrib import SIFTMatching, SURFMatching, BRIEFMatching
from airtestProject.commons.utils.debugFeature import color_block_recognition

MATCHING_METHODS = {
    "tpl": TemplateMatching,
    "mstpl": MultiScaleTemplateMatchingPre,
    "gmstpl": MultiScaleTemplateMatching,
    "kaze": KAZEMatching,
    "brisk": BRISKMatching,
    "akaze": AKAZEMatching,
    "orb": ORBMatching,
    "sift": SIFTMatching,
    "surf": SURFMatching,
    "brief": BRIEFMatching,
}


@logwrap
def loop_find(query, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, intervalfunc=None):
    """
    Search for image template in the screen until timeout

    Args:
        query: image template to be found in screenshot
        timeout: time interval how long to look for the image template
        threshold: default is None
        interval: sleep interval before next attempt to find the image template
        intervalfunc: function that is executed after unsuccessful attempt to find the image template

    Raises:
        TargetNotFoundError: when image template is not found in screenshot

    Returns:
        TargetNotFoundError if image template not found, otherwise returns the position where the image template has
        been found in screenshot

    """
    G.LOGGING.info("Try finding: %s", query)
    start_time = time.time()
    while True:
        screen = G.DEVICE.snapshot(filename=None, quality=ST.SNAPSHOT_QUALITY)

        if screen is None:
            G.LOGGING.warning("Screen is None, may be locked")
        else:
            if threshold:
                query.threshold = threshold
            match_pos = query.match_in(screen)
            if match_pos:
                try_log_screen(screen)
                return match_pos

        if intervalfunc is not None:
            intervalfunc()

        # 超时则raise，未超时则进行下次循环:
        if (time.time() - start_time) > timeout:
            try_log_screen(screen)
            raise TargetNotFoundError('Picture %s not found in screen' % query)
        else:
            time.sleep(interval)


@logwrap
def try_log_screen(screen=None, quality=None, max_size=None):
    """
    Save screenshot to file

    Args:
        screen: screenshot to be saved
        quality: The image quality, default is ST.SNAPSHOT_QUALITY
        max_size: the maximum size of the picture, e.g 1200

    Returns:
        {"screen": filename, "resolution": aircv.get_resolution(screen)}

    """
    if not ST.LOG_DIR or not ST.SAVE_IMAGE:
        return
    if not quality:
        quality = ST.SNAPSHOT_QUALITY
    if not max_size:
        max_size = ST.IMAGE_MAXSIZE
    if screen is None:
        screen = G.DEVICE.snapshot(quality=quality)
    screen = color_block_recognition(screen, lowerb_list=[(108, 142, 66)],
                                     upperb_list=[(119, 255, 255),])
    filename = "%(time)d.jpg" % {'time': time.time() * 1000}
    filepath = os.path.join(ST.LOG_DIR, filename)
    if screen is not None:
        aircv.imwrite(filepath, screen, quality, max_size=max_size)
        return {"screen": filename, "resolution": aircv.get_resolution(screen)}
    return None


class Template(object):
    """
    picture as touch/swipe/wait/exists target and extra info for cv match
    filename: pic filename
    target_pos: ret which pos in the pic
    record_pos: pos in screen when recording
    resolution: screen resolution when recording
    rgb: 识别结果是否使用rgb三通道进行校验.
    scale_max: 多尺度模板匹配最大范围.
    scale_step: 多尺度模板匹配搜索步长.
    """

    def __init__(self, filename, threshold=None, target_pos=TargetPos.MID, record_pos=None, resolution=(), rgb=False, scale_max=800, scale_step=0.005):
        self.filename = filename
        self._filepath = None
        self.threshold = threshold or ST.THRESHOLD
        self.target_pos = target_pos
        self.record_pos = record_pos
        self.resolution = resolution
        self.rgb = rgb
        self.scale_max = scale_max
        self.scale_step = scale_step

    @property
    def filepath(self):
        if self._filepath:
            return self._filepath
        for dirname in G.BASEDIR:
            filepath = os.path.join(dirname, self.filename)
            if os.path.isfile(filepath):
                self._filepath = filepath
                return self._filepath
        return self.filename

    def __repr__(self):
        filepath = self.filepath if PY3 else self.filepath.encode(sys.getfilesystemencoding())
        return "Template(%s)" % filepath

    def match_in(self, screen):
        match_result = self._cv_match(screen)
        G.LOGGING.debug("match result: %s", match_result)
        if not match_result:
            return None
        focus_pos = TargetPos().getXY(match_result, self.target_pos)
        return focus_pos

    def match_all_in(self, screen):
        image = self._imread()
        image = self._resize_image(image, screen, ST.RESIZE_METHOD)
        return self._find_all_template(image, screen)

    @logwrap
    def _cv_match(self, screen):
        # in case image file not exist in current directory:
        ori_image = self._imread()
        image = self._resize_image(ori_image, screen, ST.RESIZE_METHOD)
        ret = None
        for method in ST.CVSTRATEGY:
            # get function definition and execute:
            func = MATCHING_METHODS.get(method, None)
            if func is None:
                raise InvalidMatchingMethodError("Undefined method in CVSTRATEGY: '%s', try 'kaze'/'brisk'/'akaze'/'orb'/'surf'/'sift'/'brief' instead." % method)
            else:
                if method in ["mstpl", "gmstpl"]:
                    ret = self._try_match(func, ori_image, screen, threshold=self.threshold, rgb=self.rgb, record_pos=self.record_pos,
                                            resolution=self.resolution, scale_max=self.scale_max, scale_step=self.scale_step)
                else:
                    ret = self._try_match(func, image, screen, threshold=self.threshold, rgb=self.rgb)
            if ret:
                break
        return ret

    @staticmethod
    def _try_match(func, *args, **kwargs):
        G.LOGGING.debug("try match with %s" % func.__name__)
        try:
            ret = func(*args, **kwargs).find_best_result()
        except aircv.NoModuleError as err:
            G.LOGGING.warning("'surf'/'sift'/'brief' is in opencv-contrib module. You can use 'tpl'/'kaze'/'brisk'/'akaze'/'orb' in CVSTRATEGY, or reinstall opencv with the contrib module.")
            return None
        except aircv.BaseError as err:
            G.LOGGING.debug(repr(err))
            return None
        else:
            return ret

    def _imread(self):
        return aircv.imread(self.filepath)

    def _find_all_template(self, image, screen):
        return TemplateMatching(image, screen, threshold=self.threshold, rgb=self.rgb).find_all_results()

    def _find_keypoint_result_in_predict_area(self, func, image, screen):
        if not self.record_pos:
            return None
        # calc predict area in screen
        image_wh, screen_resolution = aircv.get_resolution(image), aircv.get_resolution(screen)
        xmin, ymin, xmax, ymax = Predictor.get_predict_area(self.record_pos, image_wh, self.resolution, screen_resolution)
        # crop predict image from screen
        predict_area = aircv.crop_image(screen, (xmin, ymin, xmax, ymax))
        if not predict_area.any():
            return None
        # keypoint matching in predicted area:
        ret_in_area = func(image, predict_area, threshold=self.threshold, rgb=self.rgb)
        # calc cv ret if found
        if not ret_in_area:
            return None
        ret = deepcopy(ret_in_area)
        if "rectangle" in ret:
            for idx, item in enumerate(ret["rectangle"]):
                ret["rectangle"][idx] = (item[0] + xmin, item[1] + ymin)
        ret["result"] = (ret_in_area["result"][0] + xmin, ret_in_area["result"][1] + ymin)
        return ret

    def _resize_image(self, image, screen, resize_method):
        """模板匹配中，将输入的截图适配成 等待模板匹配的截图."""
        # 未记录录制分辨率，跳过
        if not self.resolution:
            return image
        screen_resolution = aircv.get_resolution(screen)
        # 如果分辨率一致，则不需要进行im_search的适配:
        if tuple(self.resolution) == tuple(screen_resolution) or resize_method is None:
            return image
        if isinstance(resize_method, types.MethodType):
            resize_method = resize_method.__func__
        # 分辨率不一致则进行适配，默认使用cocos_min_strategy:
        h, w = image.shape[:2]
        w_re, h_re = resize_method(w, h, self.resolution, screen_resolution)
        # 确保w_re和h_re > 0, 至少有1个像素:
        w_re, h_re = max(1, w_re), max(1, h_re)
        # 调试代码: 输出调试信息.
        G.LOGGING.debug("resize: (%s, %s)->(%s, %s), resolution: %s=>%s" % (
                        w, h, w_re, h_re, self.resolution, screen_resolution))
        # 进行图片缩放:
        image = cv2.resize(image, (w_re, h_re))
        return image


class Predictor(object):
    """
    this class predicts the press_point and the area to search im_search.
    """

    DEVIATION = 100

    @staticmethod
    def count_record_pos(pos, resolution):
        """计算坐标对应的中点偏移值相对于分辨率的百分比."""
        _w, _h = resolution
        # 都按宽度缩放，针对G18的实验结论
        delta_x = (pos[0] - _w * 0.5) / _w
        delta_y = (pos[1] - _h * 0.5) / _w
        delta_x = round(delta_x, 3)
        delta_y = round(delta_y, 3)
        return delta_x, delta_y

    @classmethod
    def get_predict_point(cls, record_pos, screen_resolution):
        """预测缩放后的点击位置点."""
        delta_x, delta_y = record_pos
        _w, _h = screen_resolution
        target_x = delta_x * _w + _w * 0.5
        target_y = delta_y * _w + _h * 0.5
        return target_x, target_y

    @classmethod
    def get_predict_area(cls, record_pos, image_wh, image_resolution=(), screen_resolution=()):
        """Get predicted area in screen."""
        x, y = cls.get_predict_point(record_pos, screen_resolution)
        # The prediction area should depend on the image size:
        if image_resolution:
            predict_x_radius = int(image_wh[0] * screen_resolution[0] / (2 * image_resolution[0])) + cls.DEVIATION
            predict_y_radius = int(image_wh[1] * screen_resolution[1] / (2 * image_resolution[1])) + cls.DEVIATION
        else:
            predict_x_radius, predict_y_radius = int(image_wh[0] / 2) + cls.DEVIATION, int(image_wh[1] / 2) + cls.DEVIATION
        area = (x - predict_x_radius, y - predict_y_radius, x + predict_x_radius, y + predict_y_radius)
        return area



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import cv2
import numpy as np

from .error import *  # noqa
from .utils import generate_result, check_image_valid
from .cal_confidence import cal_ccoeff_confidence, cal_rgb_confidence

# SIFT识别特征点匹配，参数设置:
FLANN_INDEX_KDTREE = 0
FLANN = cv2.FlannBasedMatcher({'algorithm': FLANN_INDEX_KDTREE, 'trees': 5}, dict(checks=50))
# SIFT参数: FILTER_RATIO为SIFT优秀特征点过滤比例值(0-1范围，建议值0.4-0.6)
FILTER_RATIO = 0.59
# SIFT参数: SIFT识别时只找出一对相似特征点时的置信度(confidence)
ONE_POINT_CONFI = 0.5


def find_sift(im_source, im_search, threshold=0.8, rgb=True, good_ratio=FILTER_RATIO):
    """基于sift进行图像识别，只筛选出最优区域."""
    # 第一步：检验图像是否正常：
    if not check_image_valid(im_source, im_search):
        return None

    # 第二步：获取特征点集并匹配出特征点对: 返回值 good, pypts, kp_sch, kp_src
    kp_sch, kp_src, good = _get_key_points(im_source, im_search, good_ratio)

    # 第三步：根据匹配点对(good),提取出来识别区域:
    if len(good) == 0:
        # 匹配点对为0,无法提取识别区域：
        return None
    elif len(good) == 1:
        # 匹配点对为1，可信度赋予设定值，并直接返回:
        return _handle_one_good_points(kp_src, good, threshold) if ONE_POINT_CONFI >= threshold else None
    elif len(good) == 2:
        # 匹配点对为2，根据点对求出目标区域，据此算出可信度：
        origin_result = _handle_two_good_points(im_source, im_search, kp_src, kp_sch, good)
        if isinstance(origin_result, dict):
            return origin_result if ONE_POINT_CONFI >= threshold else None
        else:
            middle_point, pypts, w_h_range = _handle_two_good_points(im_source, im_search, kp_src, kp_sch, good)
    elif len(good) == 3:
        # 匹配点对为3，取出点对，求出目标区域，据此算出可信度：
        origin_result = _handle_three_good_points(im_source, im_search, kp_src, kp_sch, good)
        if isinstance(origin_result, dict):
            return origin_result if ONE_POINT_CONFI >= threshold else None
        else:
            middle_point, pypts, w_h_range = _handle_three_good_points(im_source, im_search, kp_src, kp_sch, good)
    else:
        # 匹配点对 >= 4个，使用单矩阵映射求出目标区域，据此算出可信度：
        middle_point, pypts, w_h_range = _many_good_pts(im_source, im_search, kp_sch, kp_src, good)

    # 第四步：根据识别区域，求出结果可信度，并将结果进行返回:
    # 对识别结果进行合理性校验: 小于5个像素的，或者缩放超过5倍的，一律视为不合法直接raise.
    _target_error_check(w_h_range)
    # 将截图和识别结果缩放到大小一致,准备计算可信度
    x_min, x_max, y_min, y_max, w, h = w_h_range
    target_img = im_source[y_min:y_max, x_min:x_max]
    resize_img = cv2.resize(target_img, (w, h))
    confidence = _cal_sift_confidence(im_search, resize_img, rgb=rgb)

    best_match = generate_result(middle_point, pypts, confidence)
    print("[aircv][sift] threshold=%s, result=%s" % (threshold, best_match))
    return best_match if confidence >= threshold else None


def mask_sift(im_source, im_search, threshold=0.8, rgb=True, good_ratio=FILTER_RATIO):
    """基于sift查找多个目标区域的方法."""
    # 求出特征点后，im_source中获得match的那些点进行聚类
    raise NotImplementedError


def find_all_sift(im_source, im_search, threshold=0.8, rgb=True, good_ratio=FILTER_RATIO):
    """基于sift查找多个目标区域的方法."""
    # 求出特征点后，im_source中获得match的那些点进行聚类
    raise NotImplementedError


def _init_sift():
    """Make sure that there is SIFT module in OpenCV."""
    if cv2.__version__.startswith("3."):
        # OpenCV3.x, sift is in contrib module, you need to compile it seperately.
        try:
            sift = cv2.xfeatures2d.SIFT_create(edgeThreshold=10)
        except:
            print("to use SIFT, you should build contrib with opencv3.0")
            raise NoSIFTModuleError("There is no SIFT module in your OpenCV environment !")
    else:
        # OpenCV2.x, just use it.
        sift = cv2.SIFT(edgeThreshold=10)

    return sift


def _get_key_points(im_source, im_search, good_ratio):
    """根据传入图像,计算图像所有的特征点,并得到匹配特征点对."""
    # 准备工作: 初始化sift算子
    sift = _init_sift()
    # 第一步：获取特征点集，并匹配出特征点对: 返回值 good, pypts, kp_sch, kp_src
    kp_sch, des_sch = sift.detectAndCompute(im_search, None)
    kp_src, des_src = sift.detectAndCompute(im_source, None)
    # When apply knnmatch , make sure that number of features in both test and
    #       query image is greater than or equal to number of nearest neighbors in knn match.
    if len(kp_sch) < 2 or len(kp_src) < 2:
        raise NoSiftMatchPointError("Not enough feature points in input images !")

    # 匹配两个图片中的特征点集，k=2表示每个特征点取出2个最匹配的对应点:
    matches = FLANN.knnMatch(des_sch, des_src, k=2)
    good = []
    # good为特征点初选结果，剔除掉前两名匹配太接近的特征点，不是独特优秀的特征点直接筛除(多目标识别情况直接不适用)
    for m, n in matches:
        if m.distance < good_ratio * n.distance:
            good.append(m)
    # good点需要去除重复的部分，（设定源图像不能有重复点）去重时将src图像中的重复点找出即可
    # 去重策略：允许搜索图像对源图像的特征点映射一对多，不允许多对一重复（即不能源图像上一个点对应搜索图像的多个点）
    good_diff, diff_good_point = [], [[]]
    for m in good:
        diff_point = [int(kp_src[m.trainIdx].pt[0]), int(kp_src[m.trainIdx].pt[1])]
        if diff_point not in diff_good_point:
            good_diff.append(m)
            diff_good_point.append(diff_point)
    good = good_diff

    return kp_sch, kp_src, good


def _handle_one_good_points(kp_src, good, threshold):
    """sift匹配中只有一对匹配的特征点对的情况."""
    # 识别中心即为该匹配点位置:
    middle_point = int(kp_src[good[0].trainIdx].pt[0]), int(kp_src[good[0].trainIdx].pt[1])
    confidence = ONE_POINT_CONFI
    # 单个特征点对,识别区域无效化:
    pypts = [middle_point for i in range(4)]
    result = generate_result(middle_point, pypts, confidence)

    return None if confidence < threshold else result


def _handle_two_good_points(im_source, im_search, kp_src, kp_sch, good):
    """处理两对特征点的情况."""
    pts_sch1 = int(kp_sch[good[0].queryIdx].pt[0]), int(kp_sch[good[0].queryIdx].pt[1])
    pts_sch2 = int(kp_sch[good[1].queryIdx].pt[0]), int(kp_sch[good[1].queryIdx].pt[1])
    pts_src1 = int(kp_src[good[0].trainIdx].pt[0]), int(kp_src[good[0].trainIdx].pt[1])
    pts_src2 = int(kp_src[good[1].trainIdx].pt[0]), int(kp_src[good[1].trainIdx].pt[1])

    return _two_good_points(pts_sch1, pts_sch2, pts_src1, pts_src2, im_search, im_source)


def _handle_three_good_points(im_source, im_search, kp_src, kp_sch, good):
    """处理三对特征点的情况."""
    # 拿出sch和src的两个点(点1)和(点2点3的中点)，
    # 然后根据两个点原则进行后处理(注意ke_sch和kp_src以及queryIdx和trainIdx):
    pts_sch1 = int(kp_sch[good[0].queryIdx].pt[0]), int(kp_sch[good[0].queryIdx].pt[1])
    pts_sch2 = int((kp_sch[good[1].queryIdx].pt[0] + kp_sch[good[2].queryIdx].pt[0]) / 2), int(
        (kp_sch[good[1].queryIdx].pt[1] + kp_sch[good[2].queryIdx].pt[1]) / 2)
    pts_src1 = int(kp_src[good[0].trainIdx].pt[0]), int(kp_src[good[0].trainIdx].pt[1])
    pts_src2 = int((kp_src[good[1].trainIdx].pt[0] + kp_src[good[2].trainIdx].pt[0]) / 2), int(
        (kp_src[good[1].trainIdx].pt[1] + kp_src[good[2].trainIdx].pt[1]) / 2)
    return _two_good_points(pts_sch1, pts_sch2, pts_src1, pts_src2, im_search, im_source)


def _many_good_pts(im_source, im_search, kp_sch, kp_src, good):
    """特征点匹配点对数目>=4个，可使用单矩阵映射,求出识别的目标区域."""
    sch_pts, img_pts = np.float32([kp_sch[m.queryIdx].pt for m in good]).reshape(
        -1, 1, 2), np.float32([kp_src[m.trainIdx].pt for m in good]).reshape(-1, 1, 2)
    # M是转化矩阵
    M, mask = _find_homography(sch_pts, img_pts)
    matches_mask = mask.ravel().tolist()
    # 从good中间筛选出更精确的点(假设good中大部分点为正确的，由ratio=0.7保障)
    selected = [v for k, v in enumerate(good) if matches_mask[k]]

    # 针对所有的selected点再次计算出更精确的转化矩阵M来
    sch_pts, img_pts = np.float32([kp_sch[m.queryIdx].pt for m in selected]).reshape(
        -1, 1, 2), np.float32([kp_src[m.trainIdx].pt for m in selected]).reshape(-1, 1, 2)
    M, mask = _find_homography(sch_pts, img_pts)
    # 计算四个角矩阵变换后的坐标，也就是在大图中的目标区域的顶点坐标:
    h, w = im_search.shape[:2]
    h_s, w_s = im_source.shape[:2]
    pts = np.float32([[0, 0], [0, h - 1], [w - 1, h - 1], [w - 1, 0]]).reshape(-1, 1, 2)
    dst = cv2.perspectiveTransform(pts, M)

    # trans numpy arrary to python list: [(a, b), (a1, b1), ...]
    def cal_rect_pts(dst):
        return [tuple(npt[0]) for npt in dst.astype(int).tolist()]

    pypts = cal_rect_pts(dst)
    # 注意：虽然4个角点有可能越出source图边界，但是(根据精确化映射单映射矩阵M线性机制)中点不会越出边界
    lt, br = pypts[0], pypts[2]
    middle_point = int((lt[0] + br[0]) / 2), int((lt[1] + br[1]) / 2)
    # 考虑到算出的目标矩阵有可能是翻转的情况，必须进行一次处理，确保映射后的“左上角”在图片中也是左上角点：
    x_min, x_max = min(lt[0], br[0]), max(lt[0], br[0])
    y_min, y_max = min(lt[1], br[1]), max(lt[1], br[1])
    # 挑选出目标矩形区域可能会有越界情况，越界时直接将其置为边界：
    # 超出左边界取0，超出右边界取w_s-1，超出下边界取0，超出上边界取h_s-1
    # 当x_min小于0时，取0。  x_max小于0时，取0。
    x_min, x_max = int(max(x_min, 0)), int(max(x_max, 0))
    # 当x_min大于w_s时，取值w_s-1。  x_max大于w_s-1时，取w_s-1。
    x_min, x_max = int(min(x_min, w_s - 1)), int(min(x_max, w_s - 1))
    # 当y_min小于0时，取0。  y_max小于0时，取0。
    y_min, y_max = int(max(y_min, 0)), int(max(y_max, 0))
    # 当y_min大于h_s时，取值h_s-1。  y_max大于h_s-1时，取h_s-1。
    y_min, y_max = int(min(y_min, h_s - 1)), int(min(y_max, h_s - 1))
    # 目标区域的角点，按左上、左下、右下、右上点序：(x_min,y_min)(x_min,y_max)(x_max,y_max)(x_max,y_min)
    pts = np.float32([[x_min, y_min], [x_min, y_max], [
                     x_max, y_max], [x_max, y_min]]).reshape(-1, 1, 2)
    pypts = cal_rect_pts(pts)

    return middle_point, pypts, [x_min, x_max, y_min, y_max, w, h]


def _two_good_points(pts_sch1, pts_sch2, pts_src1, pts_src2, im_search, im_source):
    """返回两对匹配特征点情形下的识别结果."""
    # 先算出中心点(在im_source中的坐标)：
    middle_point = [int((pts_src1[0] + pts_src2[0]) / 2), int((pts_src1[1] + pts_src2[1]) / 2)]
    pypts = []
    # 如果特征点同x轴或同y轴(无论src还是sch中)，均不能计算出目标矩形区域来，此时返回值同good=1情形
    if pts_sch1[0] == pts_sch2[0] or pts_sch1[1] == pts_sch2[1] or pts_src1[0] == pts_src2[0] or pts_src1[1] == pts_src2[1]:
        confidence = ONE_POINT_CONFI
        one_match = generate_result(middle_point, pypts, confidence)
        return one_match
    # 计算x,y轴的缩放比例：x_scale、y_scale，从middle点扩张出目标区域:(注意整数计算要转成浮点数结果!)
    h, w = im_search.shape[:2]
    h_s, w_s = im_source.shape[:2]
    x_scale = abs(1.0 * (pts_src2[0] - pts_src1[0]) / (pts_sch2[0] - pts_sch1[0]))
    y_scale = abs(1.0 * (pts_src2[1] - pts_src1[1]) / (pts_sch2[1] - pts_sch1[1]))
    # 得到scale后需要对middle_point进行校正，并非特征点中点，而是映射矩阵的中点。
    sch_middle_point = int((pts_sch1[0] + pts_sch2[0]) / 2), int((pts_sch1[1] + pts_sch2[1]) / 2)
    middle_point[0] = middle_point[0] - int((sch_middle_point[0] - w / 2) * x_scale)
    middle_point[1] = middle_point[1] - int((sch_middle_point[1] - h / 2) * y_scale)
    middle_point[0] = max(middle_point[0], 0)  # 超出左边界取0  (图像左上角坐标为0,0)
    middle_point[0] = min(middle_point[0], w_s - 1)  # 超出右边界取w_s-1
    middle_point[1] = max(middle_point[1], 0)  # 超出上边界取0
    middle_point[1] = min(middle_point[1], h_s - 1)  # 超出下边界取h_s-1

    # 计算出来rectangle角点的顺序：左上角->左下角->右下角->右上角， 注意：暂不考虑图片转动
    # 超出左边界取0, 超出右边界取w_s-1, 超出下边界取0, 超出上边界取h_s-1
    x_min, x_max = int(max(middle_point[0] - (w * x_scale) / 2, 0)), int(
        min(middle_point[0] + (w * x_scale) / 2, w_s - 1))
    y_min, y_max = int(max(middle_point[1] - (h * y_scale) / 2, 0)), int(
        min(middle_point[1] + (h * y_scale) / 2, h_s - 1))
    # 目标矩形的角点按左上、左下、右下、右上的点序：(x_min,y_min)(x_min,y_max)(x_max,y_max)(x_max,y_min)
    pts = np.float32([[x_min, y_min], [x_min, y_max], [x_max, y_max], [x_max, y_min]]).reshape(-1, 1, 2)
    for npt in pts.astype(int).tolist():
        pypts.append(tuple(npt[0]))

    return middle_point, pypts, [x_min, x_max, y_min, y_max, w, h]


def _find_homography(sch_pts, src_pts):
    """多组特征点对时，求取单向性矩阵."""
    try:
        M, mask = cv2.findHomography(sch_pts, src_pts, cv2.RANSAC, 5.0)
    except Exception:
        import traceback
        traceback.print_exc()
        raise HomographyError("OpenCV error in _find_homography()...")
    else:
        if mask is None:
            raise HomographyError("In _find_homography(), find no mask...")
        else:
            return M, mask


def _target_error_check(w_h_range):
    """校验识别结果区域是否符合常理."""
    x_min, x_max, y_min, y_max, w, h = w_h_range
    tar_width, tar_height = x_max - x_min, y_max - y_min
    # 如果src_img中的矩形识别区域的宽和高的像素数＜5，则判定识别失效。认为提取区域待不可能小于5个像素。(截图一般不可能小于5像素)
    if tar_width < 5 or tar_height < 5:
        raise SiftResultCheckError("In src_image, Taget area: width or height < 5 pixel.")
    # 如果矩形识别区域的宽和高，与sch_img的宽高差距超过5倍(屏幕像素差不可能有5倍)，认定为识别错误。
    if tar_width < 0.2 * w or tar_width > 5 * w or tar_height < 0.2 * h or tar_height > 5 * h:
        raise SiftResultCheckError("Target area is 5 times bigger or 0.2 times smaller than sch_img.")


def _cal_sift_confidence(im_search, resize_img, rgb=False):
    if rgb:
        confidence = cal_rgb_confidence(resize_img, im_search)
    else:
        confidence = cal_ccoeff_confidence(resize_img, im_search)
    # sift的confidence要放水
    confidence = (1 + confidence) / 2
    return confidence



# _*_ coding:UTF-8 _*_


class TargetPos(object):
    """
    点击目标图片的不同位置，默认为中心点0
    1 2 3
    4 0 6
    7 8 9
    """
    LEFTUP, UP, RIGHTUP = 1, 2, 3
    LEFT, MID, RIGHT = 4, 5, 6
    LEFTDOWN, DOWN, RIGHTDOWN = 7, 8, 9

    def getXY(self, cvret, pos):
        if pos == 0 or pos == self.MID:
            return cvret["result"]
        rect = cvret.get("rectangle")
        if not rect:
            print("could not get rectangle, use mid point instead")
            return cvret["result"]
        w = rect[2][0] - rect[0][0]
        h = rect[2][1] - rect[0][1]
        if pos == self.LEFTUP:
            return rect[0]
        elif pos == self.LEFTDOWN:
            return rect[1]
        elif pos == self.RIGHTDOWN:
            return rect[2]
        elif pos == self.RIGHTUP:
            return rect[3]
        elif pos == self.LEFT:
            return rect[0][0], rect[0][1] + h / 2
        elif pos == self.UP:
            return rect[0][0] + w / 2, rect[0][1]
        elif pos == self.RIGHT:
            return rect[2][0], rect[2][1] - h / 2
        elif pos == self.DOWN:
            return rect[2][0] - w / 2, rect[2][1]
        else:
            print("invalid target_pos:%s, use mid point instead" % pos)
            return cvret["result"]



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Use python to forward the iOS mobile phone port
# refer to: http://iphonedevwiki.net/index.php/SSH_Over_USB
# https://github.com/nabla-c0d3/multcprelay
# 通过Python对iOS端口进行转发，可以参考网上的tcprelay代码，或是tidevice的relay.py（使用了tornado）
import socketserver as SocketServer
import select
import sys
import threading
from optparse import OptionParser
from airtestProject.airtest.utils.logger import get_logger


LOGGING = get_logger(__name__)


class SocketRelay(object):
    def __init__(self, a, b, maxbuf=65535):
        self.a = a
        self.b = b
        self.atob = b""
        self.btoa = b""
        self.maxbuf = maxbuf

    def handle(self):
        while True:
            rlist = []
            wlist = []
            xlist = [self.a, self.b]
            if self.atob:
                wlist.append(self.b)
            if self.btoa:
                wlist.append(self.a)
            if len(self.atob) < self.maxbuf:
                rlist.append(self.a)
            if len(self.btoa) < self.maxbuf:
                rlist.append(self.b)
            rlo, wlo, xlo = select.select(rlist, wlist, xlist)
            if xlo:
                return
            if self.a in wlo:
                n = self.a.send(self.btoa)
                self.btoa = self.btoa[n:]
            if self.b in wlo:
                n = self.b.send(self.atob)
                self.atob = self.atob[n:]
            if self.a in rlo:
                s = self.a.recv(self.maxbuf - len(self.atob))
                if not s:
                    return
                self.atob += s
            if self.b in rlo:
                s = self.b.recv(self.maxbuf - len(self.btoa))
                if not s:
                    return
                self.btoa += s


class TCPRelay(SocketServer.BaseRequestHandler):
    def handle(self):
        dev = self.server.device
        dsock = dev.create_inner_connection(self.server.rport)._sock
        lsock = self.request
        LOGGING.info("Connection established, relaying data")
        try:
            fwd = SocketRelay(dsock, lsock, self.server.bufsize * 1024)
            fwd.handle()
        finally:
            dsock.close()
            lsock.close()
        LOGGING.info("Connection closed")


class TCPServer(SocketServer.TCPServer):
    allow_reuse_address = True


class ThreadedTCPServer(SocketServer.ThreadingMixIn, TCPServer):
    # 显式指定为True，否则脚本运行完毕时，因为连接没有断开，导致线程不会终止
    daemon_threads = True


if __name__ == '__main__':
    """
    本文件可以在usb仅插入一台iOS手机时，执行命令行：python relay.py -t 5001:5001
    """
    from wda.usbmux import Usbmux
    HOST = "localhost"

    parser = OptionParser(usage="usage: %prog [OPTIONS] RemotePort[:LocalPort] [RemotePort[:LocalPort]]...")
    parser.add_option("-t", "--threaded", dest='threaded', action='store_true', default=False, help="use threading to handle multiple connections at once")
    parser.add_option("-b", "--bufsize", dest='bufsize', action='store', metavar='KILOBYTES', type='int', default=128, help="specify buffer size for socket forwarding")
    parser.add_option("-s", "--socket", dest='sockpath', action='store', metavar='PATH', type='str', default=None, help="specify the path of the usbmuxd socket")

    options, args = parser.parse_args()

    serverclass = TCPServer
    if options.threaded:
        serverclass = ThreadedTCPServer

    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    ports = []

    for arg in args:
        try:
            if ':' in arg:
                rport, lport = arg.split(":")
                rport = int(rport)
                lport = int(lport)
                ports.append((rport, lport))
            else:
                ports.append((int(arg), int(arg)))
        except:
            parser.print_help()
            sys.exit(1)

    servers=[]

    for rport, lport in ports:
        LOGGING.info("Forwarding local port %d to remote port %d"%(lport, rport))
        server = serverclass((HOST, lport), TCPRelay)
        # 当前仅有一台iOS手机连接usb时
        dev_uuid = Usbmux().get_single_device_udid()
        dev = Usbmux().device(dev_uuid)
        server.rport = rport
        server.device = dev
        server.bufsize = options.bufsize
        servers.append(server)

    alive = True

    while alive:
        # sleep(1)
        try:
            rl, wl, xl = select.select(servers, [], [])
            for server in rl:
                server.handle_request()
        except:
            alive = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import cv2
import time
import numpy as np
from PIL import Image

from airtestProject.airtest.utils.logger import get_logger
from .error import TemplateInputError

LOGGING = get_logger(__name__)


def print_run_time(func):

    def wrapper(self, *args, **kwargs):
        start_time = time.time()
        ret = func(self, *args, **kwargs)
        t = time.time() - start_time
        LOGGING.debug("%s() run time is %.2f s." % (func.__name__, t))
        if ret and isinstance(ret, dict):
            ret["time"] = t
        return ret

    return wrapper


def generate_result(middle_point, pypts, confi):
    """Format the result: 定义图像识别结果格式."""
    ret = dict(result=middle_point,
               rectangle=pypts,
               confidence=confi)
    return ret


def check_image_valid(im_source, im_search):
    """Check if the input images valid or not."""
    if im_source is not None and im_source.any() and im_search is not None and im_search.any():
        return True
    else:
        return False


def check_source_larger_than_search(im_source, im_search):
    """检查图像识别的输入."""
    # 图像格式, 确保输入图像为指定的矩阵格式:
    # 图像大小, 检查截图宽、高是否大于了截屏的宽、高:
    h_search, w_search = im_search.shape[:2]
    h_source, w_source = im_source.shape[:2]
    if h_search > h_source or w_search > w_source:
        raise TemplateInputError("error: in template match, found im_search bigger than im_source.")


def img_mat_rgb_2_gray(img_mat):
    """
    Turn img_mat into gray_scale, so that template match can figure the img data.
    "print(type(im_search[0][0])")  can check the pixel type.
    """
    assert isinstance(img_mat[0][0], np.ndarray), "input must be instance of np.ndarray"
    return cv2.cvtColor(img_mat, cv2.COLOR_BGR2GRAY)


def img_2_string(img):
    _, png = cv2.imencode('.png', img)
    return png.tostring()


def string_2_img(pngstr):
    nparr = np.frombuffer(pngstr, np.uint8)
    img = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    return img


def pil_2_cv2(pil_image):
    open_cv_image = np.array(pil_image)
    # Convert RGB to BGR (method-1):
    open_cv_image = cv2.cvtColor(open_cv_image, cv2.COLOR_RGB2BGR)
    # Convert RGB to BGR (method-2):
    # b, g, r = cv2.split(open_cv_image)
    # open_cv_image = cv2.merge([r, g, b])
    return open_cv_image


def cv2_2_pil(cv2_image):
    cv2_im = cv2.cvtColor(cv2_image, cv2.COLOR_BGR2RGB)
    pil_im = Image.fromarray(cv2_im)
    return pil_im


def compress_image(pil_img, path, quality, max_size=None):
    """
    Save the picture and compress

    :param pil_img: PIL image
    :param path: save path
    :param quality: the image quality, integer in range [1, 99]
    :param max_size: the maximum size of the picture, e.g 1200
    :return:
    """
    if max_size:
        # The picture will be saved in a size <= max_size*max_size
        pil_img.thumbnail((max_size, max_size), Image.LANCZOS)
    quality = int(round(quality))
    if quality <= 0 or quality >= 100:
        raise Exception("SNAPSHOT_QUALITY (" + str(quality) + ") should be an integer in the range [1,99]")
    pil_img.save(path, quality=quality, optimize=True)



import logging
import os
import time
import traceback

import six
from airtestProject.airtest.aircv import aircv
from airtestProject.airtest.core.helper import G
from airtestProject.airtest.utils.logger import get_logger
from airtestProject.airtest.utils.logwraper import AirtestLogger
from loguru import logger
from functools import wraps

# 设置airtest日志等级，debug会输出很多日志
air_logger = logging.getLogger("airtest")
air_logger.setLevel(logging.ERROR)


class LogManager():
    def __init__(self, logfile, log_file="all.txt", project=None):
        self.LOGGER = AirtestLogger(logfile)
        if project is not None:
            self.log_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__))) + f"\\log\\{project}"
            self.log_file = os.path.join(self.log_dir, log_file)
            self.setup_logger()
            self.handler = None
        self.save_image = False
        self.project = project

    def setup_logger(self):
        logger.remove()
        self.handler = logger.add(self.log_file, rotation="10 MB", backtrace=True, diagnose=True)

    def close_logger(self):
        if self.handler is not None:
            logger.remove(self.handler)
            self.handler = None

    def log_step(self, message):
        """
        步骤日志，用来标记测试用例中每个步骤的记录日志，日志级别是info级别，便于问题跟踪与定位
        :param message:日志消息内容
        :return:None
        """
        self.log(arg=message, timestamp=time.time(), snapshot=True)
        logger.info(message)

    def log_error(self, message):
        """
        错误日志，用来标记报错时的日志，日志等级是error级别
        :param message: 日志消息内容
        :return:
        """
        self.log(arg=message, timestamp=time.time(), snapshot=False)
        logger.error(message)

    @staticmethod
    def log_assertion(assertion, expected, actual):
        """
        断言日志，用例标记断言过程的日志，日志等级为info级别
        :param assertion:true/flase,表示是否需要断言的标记
        :param expected:期望值
        :param actual:实际值
        :return:无返回
        """
        if assertion:
            logger.info(f"断言通过: 期望值 - {expected}, 实际值 - {actual}")
        else:
            logger.error(f"断言失败: 期望值 - {expected}, 实际值 - {actual}")

    def log_subprocess(self, function_name, *args, **kwargs):
        """
        记录执行子流程的日志，用于跟踪自动化执行过程中的子流程信息。日志等级为DEBUG级别。
        :param function_name: str，子流程的函数名。
        :param args: tuple，传递给子流程函数的位置参数。
        :param kwargs: dict，传递给子流程函数的关键字参数。
        :return: None
        """
        self.log(arg=f"执行子流程: {function_name} 使用参数: {args}, kwargs: {kwargs}", timestamp=time.time())
        logger.debug(f"执行子流程: {function_name} 使用参数: {args}, kwargs: {kwargs}")

    def log(self, arg, timestamp=None, desc="", snapshot=False):
        """
        记录日志
        :param arg: log信息或者异常对象。
        :param snapshot: 是否截图。
        :param desc: int，日志描述、默认是arg.class.__name__。
        :param timestamp: 默认是time.time()，记录时间戳。
        """
        depth = 0
        screen_data = None
        if snapshot:
            # 如果指定了snapshot参数，强制保存一张图片
            self.save_image = True
            try:
                screen_data = self.try_log_screen()
                depth = 2
            except AttributeError:
                # if G.DEVICE is None
                pass
            else:
                depth = 1
            finally:
                self.save_image = False
        if isinstance(arg, Exception):
            if hasattr(arg, "__traceback__"):
                # in PY3, arg.__traceback__ is traceback object
                trace_msg = ''.join(traceback.format_exception(type(arg), arg, arg.__traceback__))
            else:
                trace_msg = arg.message  # PY2
            AirtestLogger.log("info", {
                "name": desc or arg.__class__.__name__,
            }, depth=depth, timestamp=timestamp)
            get_logger("airtest.core.api").error(trace_msg)
        elif isinstance(arg, six.string_types):
            # 普通文本log内容放在"log"里，如果有trace内容放在"traceback"里
            # 在报告中，假如"traceback"有内容，将会被识别为报错，这个步骤会被判定为不通过
            self.LOGGER.log("info", {"name": desc or arg, "traceback": None, "ret": screen_data ,"log": arg}, depth=depth,
                            timestamp=timestamp)
            get_logger("airtest.core.api").info(arg)
        else:
            # arg['ret'] = screen_data

            self.LOGGER.log("info",
                            {"name": desc or repr(arg), "traceback": None, "log": repr(arg), "ret": screen_data},
                            depth=depth,
                            timestamp=timestamp)
            get_logger("airtest.core.api").info(repr(arg))

    def try_log_screen(self, screen=None, quality=None, max_size=None):
        """
        Save screenshot to file

        Args:
            screen: screenshot to be saved
            quality: The image quality, default is ST.SNAPSHOT_QUALITY
            max_size: the maximum size of the picture, e.g 1200

        Returns:
            {"screen": filename, "resolution": aircv.get_resolution(screen)}

        """
        if not self.log_dir or not self.save_image:
            return
        if not quality:
            quality = 10
        if not max_size:
            max_size = 600
        if screen is None:
            screen = G.DEVICE.snapshot(quality=quality)
        filename = "%(time)d.jpg" % {'time': time.time() * 1000}
        filepath = os.path.join(self.log_dir, filename)
        if screen is not None:
            aircv.imwrite(filepath, screen, quality, max_size=max_size)
            print({"screen": filename, "resolution": aircv.get_resolution(screen)})
            return {"screen": filename, "resolution": aircv.get_resolution(screen)}
        return None



def catch_error(func):
    """
    装饰器: 打印报错、通知
    """
    @wraps(func)  # 保持原函数的名称和文档字符串
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as e:

            logger.error(traceback.format_exc())
    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""These functions calculate the similarity of two images of the same size."""


import cv2
import numpy as np
from .utils import img_mat_rgb_2_gray


def cal_ccoeff_confidence(im_source, im_search):
    """求取两张图片的可信度，使用TM_CCOEFF_NORMED方法."""
    # 扩展置信度计算区域
    im_source = cv2.copyMakeBorder(im_source, 10,10,10,10,cv2.BORDER_REPLICATE)
    # 加入取值范围干扰，防止算法过于放大微小差异
    im_source[0,0] = 0
    im_source[0,1] = 255

    im_source, im_search = img_mat_rgb_2_gray(im_source), img_mat_rgb_2_gray(im_search)
    res = cv2.matchTemplate(im_source, im_search, cv2.TM_CCOEFF_NORMED)
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)

    return max_val


def cal_rgb_confidence(img_src_rgb, img_sch_rgb):
    """同大小彩图计算相似度."""
    # 减少极限值对hsv角度计算的影响
    img_src_rgb = np.clip(img_src_rgb, 10, 245)
    img_sch_rgb = np.clip(img_sch_rgb, 10, 245)
    # 转HSV强化颜色的影响
    img_src_rgb = cv2.cvtColor(img_src_rgb, cv2.COLOR_BGR2HSV)
    img_sch_rgb = cv2.cvtColor(img_sch_rgb, cv2.COLOR_BGR2HSV)

    # 扩展置信度计算区域
    img_src_rgb = cv2.copyMakeBorder(img_src_rgb, 10,10,10,10,cv2.BORDER_REPLICATE)
    # 加入取值范围干扰，防止算法过于放大微小差异
    img_src_rgb[0,0] = 0
    img_src_rgb[0,1] = 255

    # 计算BGR三通道的confidence，存入bgr_confidence
    src_bgr, sch_bgr = cv2.split(img_src_rgb), cv2.split(img_sch_rgb)
    bgr_confidence = [0, 0, 0]
    for i in range(3):
        res_temp = cv2.matchTemplate(src_bgr[i], sch_bgr[i], cv2.TM_CCOEFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res_temp)
        bgr_confidence[i] = max_val

    return min(bgr_confidence)



from airtestProject.airtest.core.cv import Template
from airtestProject.airtest.core.helper import logwrap

from airtestProject.airtest.core.settings import Settings as ST  # noqa
import cv2
from airtestProject.airtest.core.error import InvalidMatchingMethodError
from airtestProject.airtest.utils.transform import TargetPos
from distutils.version import LooseVersion

from airtestProject.airtest.aircv.template_matching import TemplateMatching
from airtestProject.airtest.aircv.multiscale_template_matching import MultiScaleTemplateMatching, MultiScaleTemplateMatchingPre
from airtestProject.airtest.aircv.keypoint_matching import KAZEMatching, BRISKMatching, AKAZEMatching, ORBMatching
from airtestProject.airtest.aircv.keypoint_matching_contrib import SIFTMatching, SURFMatching, BRIEFMatching

MATCHING_METHODS = {
    "sift": SIFTMatching,
    "orb": ORBMatching,
    "surf": SURFMatching,
    "tpl": TemplateMatching,
    "mstpl": MultiScaleTemplateMatchingPre,
    "gmstpl": MultiScaleTemplateMatching,
    "kaze": KAZEMatching,
    "brisk": BRISKMatching,
    "akaze": AKAZEMatching,
    "brief": BRIEFMatching,
}


class myTemplate(Template):
    CVSTRATEGY = ["surf", "sift", "mstpl", "tpl"]
    if LooseVersion('3.4.2') < LooseVersion(cv2.__version__) < LooseVersion('4.4.0'):
        CVSTRATEGY = ["mstpl", "tpl", "brisk"]

    def __init__(self, filename, threshold=None, target_pos=TargetPos.MID, record_pos=None, resolution=(), rgb=False,
                 scale_max=800, scale_step=0.005):
        super().__init__(filename, threshold, target_pos, record_pos, resolution, rgb, scale_max, scale_step)

    @logwrap
    def _cv_match(self, screen):
        # in case image file not exist in current directory:
        ori_image = self._imread()
        image = self._resize_image(ori_image, screen, ST.RESIZE_METHOD)
        ret = None
        for method in myTemplate.CVSTRATEGY:
            # get function definition and execute:
            func = MATCHING_METHODS.get(method, None)
            if func is None:
                raise InvalidMatchingMethodError("Undefined method in CVSTRATEGY: '%s', try "
                                                 "'kaze'/'brisk'/'akaze'/'orb'/'surf'/'sift'/'brief' instead." % method)
            else:
                if method in ["mstpl", "gmstpl"]:
                    ret = self._try_match(func, ori_image, screen, threshold=self.threshold, rgb=self.rgb,
                                          record_pos=self.record_pos,
                                          resolution=self.resolution, scale_max=self.scale_max,
                                          scale_step=self.scale_step)
                else:
                    ret = self._try_match(func, image, screen, threshold=self.threshold, rgb=self.rgb)
            if ret:
                break
        return ret



from poco.drivers.unity3d import UnityPoco
from airtestProject.commons.UWA import *
from airtestProject.airtest.core.api import *

from airtestProject.commons.utils.appStart import AppStartCase
import time
import unittest


class TestBaseCase(AppStartCase):

    def setUp(self):
        """
        开始自动化测试的准备工作：
        1、启动APP
        2、实例化poco，注意：unitypoco的实例化需要等游戏启动稳定后再实例化，否则可能会出现实例化失败的情况
        3、连接uwa接口
        4、启动uwa测试模式
        """
        print('unittest测试')
        # start_app('com.sanqi.odinUWA')
        time.sleep(10)
        self.poco = UnityPoco()
        time.sleep(5)
        GOT_Test.Connect(self.poco)
        # start需要带上mode模式，不能不带，否则上传不了，可能uwa.py的源码些得有点问题
        GOT_Test.Start(self.poco, 'overview')
        # time.sleep(5)


    def test_login(self):
        """
        用例测试过程，整体时间需要超过60s，否则uwa可能无法上传数据
        """
        time.sleep(5)
        print('开始执行点击登录按钮')
        self.poco("btnLogin").click()
        print('点击登录完毕')
        # dump
        GOT_Test.Dump(self.poco, 'overdraw')
        self.poco().wait_for_appearance()


    def tearDown(self):
        # 账号：wangmo@37.com
        # 密码：37youxi123456
        print("执行tearDown步骤")
        GOT_Test.Stop(self.poco)
        GOT_Test.LocalUpload(self.poco, account='wangmo@37.com', password='37youxi123456', projectID=1775,
                             timeLimitS=300)
        # 数据上传后，会显示一个uwa数据上传完成的确定界面，但是界面不可点击操作，因此每次自动化跑完后，需要stop掉APP
        stop_app('com.sanqi.odinUWA')


# if __name__ == '__main__':
    # unittest.suite



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""模板匹配.

对用户提供的调节参数:
    1. threshod: 筛选阈值，默认为0.8
    2. rgb: 彩色三通道,进行彩色权识别.
"""


import cv2
from airtestProject.airtest.utils.logger import get_logger
from .utils import generate_result, check_source_larger_than_search, img_mat_rgb_2_gray
from .cal_confidence import cal_rgb_confidence
LOGGING = get_logger(__name__)


def find_template(im_source, im_search, threshold=0.8, rgb=False):
    """函数功能：找到最优结果."""
    # 第一步：校验图像输入
    check_source_larger_than_search(im_source, im_search)
    # 第二步：计算模板匹配的结果矩阵res
    res = _get_template_result_matrix(im_source, im_search)
    # 第三步：依次获取匹配结果
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
    h, w = im_search.shape[:2]
    # 求取可信度:
    confidence = _get_confidence_from_matrix(im_source, im_search, max_loc, max_val, w, h, rgb)
    # 求取识别位置: 目标中心 + 目标区域:
    middle_point, rectangle = _get_target_rectangle(max_loc, w, h)
    best_match = generate_result(middle_point, rectangle, confidence)
    LOGGING.debug("threshold=%s, result=%s" % (threshold, best_match))
    return best_match if confidence >= threshold else None


def find_all_template(im_source, im_search, threshold=0.8, rgb=False, max_count=10):
    """根据输入图片和参数设置,返回所有的图像识别结果."""
    # 第一步：校验图像输入
    check_source_larger_than_search(im_source, im_search)

    # 第二步：计算模板匹配的结果矩阵res
    res = _get_template_result_matrix(im_source, im_search)

    # 第三步：依次获取匹配结果
    result = []
    h, w = im_search.shape[:2]

    while True:
        # 本次循环中,取出当前结果矩阵中的最优值
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        # 求取可信度:
        confidence = _get_confidence_from_matrix(im_source, im_search, max_loc, max_val, w, h, rgb)

        if confidence < threshold or len(result) > max_count:
            break

        # 求取识别位置: 目标中心 + 目标区域:
        middle_point, rectangle = _get_target_rectangle(max_loc, w, h)
        one_good_match = generate_result(middle_point, rectangle, confidence)

        result.append(one_good_match)

        # 屏蔽已经取出的最优结果,进入下轮循环继续寻找:
        # cv2.floodFill(res, None, max_loc, (-1000,), max(max_val, 0), flags=cv2.FLOODFILL_FIXED_RANGE)
        cv2.rectangle(res, (int(max_loc[0] - w / 2), int(max_loc[1] - h / 2)), (int(max_loc[0] + w / 2), int(max_loc[1] + h / 2)), (0, 0, 0), -1)

    return result if result else None


def _get_confidence_from_matrix(im_source, im_search, max_loc, max_val, w, h, rgb):
    """根据结果矩阵求出confidence."""
    # 求取可信度:
    if rgb:
        # 如果有颜色校验,对目标区域进行BGR三通道校验:
        img_crop = im_source[max_loc[1]:max_loc[1] + h, max_loc[0]: max_loc[0] + w]
        confidence = cal_rgb_confidence(img_crop, im_search)
    else:
        confidence = max_val

    return confidence


def _get_template_result_matrix(im_source, im_search):
    """求取模板匹配的结果矩阵."""
    # 灰度识别: cv2.matchTemplate( )只能处理灰度图片参数
    s_gray, i_gray = img_mat_rgb_2_gray(im_search), img_mat_rgb_2_gray(im_source)
    return cv2.matchTemplate(i_gray, s_gray, cv2.TM_CCOEFF_NORMED)


def _get_target_rectangle(left_top_pos, w, h):
    """根据左上角点和宽高求出目标区域."""
    x_min, y_min = left_top_pos
    # 中心位置的坐标:
    x_middle, y_middle = int(x_min + w / 2), int(y_min + h / 2)
    # 左下(min,max)->右下(max,max)->右上(max,min)
    left_bottom_pos, right_bottom_pos = (x_min, y_min + h), (x_min + w, y_min + h)
    right_top_pos = (x_min + w, y_min)
    # 点击位置:
    middle_point = (x_middle, y_middle)
    # 识别目标区域: 点序:左上->左下->右下->右上, 左上(min,min)右下(max,max)
    rectangle = (left_top_pos, left_bottom_pos, right_bottom_pos, right_top_pos)

    return middle_point, rectangle



"""
这里准备用来解析Excel表，生成py测试用例的逻辑
"""
import os.path
import time
import genaratePy

from airtestProject import config

full_case_path_list = []

def ganerate_test_case(project="zg", page=0):
    test_case_data = genaratePy.parse_excel(project=project, page=page)

    def _format_param(param):
        if isinstance(param, str):
            return f'"{param}"'
        elif isinstance(param, int):
            return param

    for key, vals in test_case_data.items():
        class_name_context = ('#!/usr/bin/python\n'
                              '# coding:utf-8\n'
                              'import unittest\n'
                              'from airtest.core.api import *\n'
                              'from poco.drivers.unity3d import UnityPoco\n'
                              'from airtestProject import config\n'
                              'from airtestProject.commons.cases.appStart import AppStartCase\n'
                              'from airtestProject.commons.utils.utils import Utils\n'
                              'import threading\n'
                              'import time\n'
                              'from airtestProject.commons.manager.LogManager import LogManager\n'
                              'log_manager = LogManager(logfile=None)\n'
                              'class ' + key + '(AppStartCase):\n'
                              )
        def_init = ('    poco = None\n'
                    '    utils = None\n'
                    '    @classmethod\n'
                    '    def setUp(cls):\n'
                    '        super().setUpClass()\n'
                    '        # log_manager.log_step(f"执行启动APP方法，包名{APP_NAME}")\n'
                    '        # start_app(APP_NAME)\n'
                    '        time.sleep(config.START_APP_TIME)\n'
                    '        log_manager.log_step("实例化poco")\n'
                    '        cls.poco = UnityPoco()\n'
                    '        cls.poco1 = None\n'
                    '        # 用于多线程任务\n'
                    '        # time.sleep(15)\n'
                    '        log_manager.log_step(f"实例化utils：{config.ZG_APP_NAME}")\n'
                    '        cls.utils = Utils(config.ZG_APP_NAME, cls.poco)\n'
                    )
        def_tear_down = ('    @classmethod\n'
                         '    def tearDownClass(cls):\n'
                         '        log_manager.log_step("执行tearDownClass方法")\n'
                         '        log_manager.log_step("执行uwa_local_upload方法")\n'
                         '        cls.utils.uwa_local_upload(config.UWA_ACCOUNT,config.UWA_PASSWORD, 3600, 600)\n'
                         )
        create_file_path = os.path.join(config.ROOT_PATH, "cases", project)
        suffix = key + '.py'
        full_case_path = os.path.join(create_file_path, suffix)
        full_case_path_list.append(full_case_path)
        with open(full_case_path, 'w', encoding='utf-8') as f:
            f.write(class_name_context)
            f.write(def_init)

        for case_key, case_vals in vals.items():
            # 测试用例方法名
            def_test_case_method = ('    def %s(self):\n'
                                    '        log_manager.log_step("执行%s方法")\n') % (case_key, case_key)
            with open(full_case_path, 'a', encoding='utf-8') as f:
                f.write(def_test_case_method)
            for operate_key, operate_value in case_vals.items():
                case_context = ''
                # if len(operateValue) > 1:
                #     if isinstance(operateValue[0], int) or isinstance(operateValue[1], int):
                #         case_context = '        self.utils.%s("%s","%s")\n' % (
                #             operateKey, operateValue[0], operateValue[1])
                # else:
                #     if isinstance(operateValue[0], int):
                #         case_context = '        self.utils.%s(%s)\n' % (operateKey, operateValue[0])
                #     else:
                #         case_context = '        self.utils.%s("%s")\n' % (operateKey, operateValue[0])

                # 暂时按照utils中每个封装的方法来实现，另外一种思路是按照参数个数来分类实现也是可以的
                if operate_key == 'uwa_connect':
                    case_context = ('        log_manager.log_step("执行%s步骤，参数%s:%s")\n'
                                    '        self.utils.%s("%s","%s")\n'
                                    '') % (
                                       operate_key, operate_value[0], operate_value[1], operate_key, operate_value[0],
                                       operate_value[1])
                elif operate_key == 'uwa_dump':
                    case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                    '        self.utils.%s("%s")\n') % (
                                       operate_key, operate_value[0], operate_key, operate_value[0])
                elif operate_key == 'uwa_tag':
                    case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                    '        self.utils.%s("%s")\n') % (
                                       operate_key, operate_value[0], operate_key, operate_value[0])
                elif operate_key == 'uwa_local_upload':
                    if len(operate_value) > 1:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s", %s)\n') % (
                                           operate_key, operate_value[0], operate_key, operate_value[0],
                                           ', '.join(map(_format_param, operate_value[1:])))
                    else:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s")\n') % (
                                           operate_key, operate_value[0], operate_key, operate_value[0])
                elif operate_key == 'poco_click':
                    if len(operate_value) > 1:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s:%s,%s")\n'
                                        '        self.utils.%s("%s","%s")\n') % (
                                           operate_key, operate_value[0], operate_value[1], operate_key,
                                           operate_value[0], operate_value[1])
                    else:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s")\n') % (
                                           operate_key, operate_value[0], operate_key, operate_value[0])
                elif operate_key == 'poco_long_click':
                    case_context = ('        log_manager.log_step("执行%s步骤，参数%s,%s")\n'
                                    '        self.utils.%s("%s",%s)\n') % (
                                       operate_key, operate_value[0], operate_value[1], operate_key, operate_value[0],
                                       operate_value[1])
                elif operate_key == 'poco_swipe':
                    if len(operate_value) > 1:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s", %s)\n') % (
                                           operate_key, operate_value[0], operate_key, operate_value[0],
                                           ', '.join(map(_format_param, operate_value[1:])))
                    else:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s")\n') % (
                                           operate_key, operate_value[0], operate_key, operate_value[0])
                elif operate_key == 'node_wait_for_appearance':
                    if len(operate_value) > 1:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s", %s)\n') % (
                                           operate_key, operate_value[0], operate_key, operate_value[0],
                                           ', '.join(map(_format_param, operate_value[1:])))
                    else:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s")\n') % (
                                           operate_key, operate_value[0], operate_key, operate_value[0])
                elif operate_key == 'node_wait_for_disappearance':
                    if len(operate_value) > 1:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s,%s")\n'
                                        '        self.utils.%s("%s",%s)\n') % (
                                           operate_key, operate_value[0], operate_value[1], operate_key,
                                           operate_value[0], operate_value[1])
                    else:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s")\n') % (
                                           operate_key, operate_value[0],
                                           operate_key, operate_value[0])
                elif operate_key == 'poco_dialog':
                    case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                    '        self.utils.%s("%s")\n') % (
                                       operate_key, operate_value[0], operate_key, operate_value[0])
                elif operate_key == 'poco_battle':
                    if len(operate_value) > 1:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s", %s)\n') % (
                                           operate_key, operate_value[0],
                                           operate_key, operate_value[0],
                                           ', '.join(map(_format_param, operate_value[1:])))
                    else:
                        case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                        '        self.utils.%s("%s")\n') % (
                                           operate_key, operate_value[0], operate_key, operate_value[0])
                elif operate_key == 'poco_use_skills':
                    case_context = ('        log_manager.log_step("执行%s步骤，参数%s,%s,%s")\n'
                                    '        self.utils.%s("%s","%s","%s")\n') % (
                                       operate_key, operate_value[0], operate_value[1], operate_value[2], operate_key,
                                       operate_value[0], operate_value[1], operate_value[2])
                elif operate_key == 'sleep':
                    case_context = ('        log_manager.log_step("执行%s步骤，参数%s")\n'
                                    '        self.utils.%s(%s)\n') % (
                                       operate_key, operate_value[0], operate_key, operate_value[0])
                with open(full_case_path, 'a', encoding='utf-8') as f:
                    f.write(case_context)

        with open(full_case_path, 'a', encoding='utf-8') as f:
            f.write(def_tear_down)


def format_code_with_autopep8(file_path_list):
    import autopep8
    for file_path in file_path_list:
        print(file_path)
        with open(file_path, 'r+', encoding='utf-8') as f:
            source_code = f.read()
            formatted_code = autopep8.fix_code(source_code)
            f.seek(0)
            f.write(formatted_code)
            f.truncate()


if __name__ == '__main__':
    ganerate_test_case()
    format_code_with_autopep8(full_case_path_list)



from abc import ABC, abstractmethod

"""
状态机状态类
"""


class State(ABC):
    @abstractmethod
    def run(self, test_case):
        pass

    def next_state(self):
        pass


class StartTest(State):
    def run(self, test_case):
        print("开始测试")
        test_case.run()

    def next_state(self):
        return TaskNormal()


class TaskException(State):
    def run(self, test_case):
        print("测试异常")

    def next_state(self):
        return TaskEnd()


class TaskNormal(State):
    def run(self, test_case):
        print("测试正常")

    def next_state(self):
        return TaskEnd()


class TaskEnd(State):
    def run(self, test_case):
        print("测试结束")

    def next_state(self):
        return self



# -*- coding: utf-8 -*-
import os

from airtestProject.airtest.core.android.constant import MAXTOUCH_JAR
from airtestProject.airtest.core.android.touch_methods.base_touch import BaseTouch
from airtestProject.airtest.utils.logger import get_logger
from airtestProject.airtest.utils.nbsp import NonBlockingStreamReader
from airtestProject.airtest.utils.safesocket import SafeSocket
from airtestProject.airtest.utils.snippet import kill_proc

LOGGING = get_logger(__name__)


class Maxtouch(BaseTouch):

    def __init__(self, adb, backend=False, size_info=None, input_event=None):
        super(Maxtouch, self).__init__(adb, backend, size_info, input_event)
        self.default_pressure = 0.5
        self.path_in_android = "/data/local/tmp/maxpresent.jar"
        self.localport = None

    def install(self):
        """
        Install maxtouch

        Returns:
            None

        """
        try:
            exists_file = self.adb.file_size(self.path_in_android)
        except:
            pass
        else:
            local_minitouch_size = int(os.path.getsize(MAXTOUCH_JAR))
            if exists_file and exists_file == local_minitouch_size:
                LOGGING.debug("install_maxtouch skipped")
                return
            self.uninstall()

        self.adb.push(MAXTOUCH_JAR, self.path_in_android)
        self.adb.shell("chmod 755 %s" % self.path_in_android)
        LOGGING.info("install maxpresent.jar finished")

    def uninstall(self):
        """
        Uninstall maxtouch

        Returns:
            None

        """
        self.adb.raw_shell("rm %s" % self.path_in_android)

    def setup_server(self):
        """
        Setup maxtouch server and adb forward

        Returns:
            server process

        """
        if self.server_proc:
            self.server_proc.kill()
            self.server_proc = None

        self.localport, deviceport = self.adb.setup_forward("localabstract:maxpresent_{}".format)
        deviceport = deviceport[len("localabstract:"):]
        p = self.adb.start_shell("app_process -Djava.class.path={0} /data/local/tmp com.netease.maxpresent.MaxPresent socket {1}".format(self.path_in_android, deviceport))

        nbsp = NonBlockingStreamReader(p.stdout, name="airtouch_server", auto_kill=True)
        line = nbsp.readline(timeout=5.0)
        if line is None:
            kill_proc(p)
            raise RuntimeError("airtouch setup timeout")

        if p.poll() is not None:
            # server setup error, may be already setup by others
            # subprocess exit immediately
            kill_proc(p)
            raise RuntimeError("airtouch server quit immediately")
        self.server_proc = p
        return p

    def setup_client(self):
        """
        Setup client

        Returns:
            None

        """
        s = SafeSocket()
        s.connect((self.adb.host, self.localport))
        s.sock.settimeout(2)
        self.client = s
        self.handle = self.safe_send

    def transform_xy(self, x, y):
        """
        Normalized coordinates (x, y)

        Args:
            x: coordinate x
            y: coordinate y

        Returns:
            transformed coordinates (x, y)

        """
        width, height = self.size_info['width'], self.size_info['height']
        return x / width, y / height

    def teardown(self):
        super(Maxtouch, self).teardown()
        if self.localport:
            self.adb.remove_forward("tcp:{}".format(self.localport))
            self.localport = None



# coding=utf-8
from __future__ import absolute_import

from airtestProject.airtest.core.device import Device
from airtestProject.airtest.core.api import connect_device, device as current_device
from airtestProject.airtest.core.error import NoDeviceError


class VirtualDevice(Device):
    def __init__(self, ip):
        super(VirtualDevice, self).__init__()
        self.ip = ip

    @property
    def uuid(self):
        return 'virtual-device'

    def get_current_resolution(self):
        return [1920, 1080]

    def get_ip_address(self):
        return self.ip


def default_device():
    """
    Get default device, if no device connected, connect to first android device.

    :return:
    """
    try:
        return current_device()
    except NoDeviceError:
        return connect_device('Android:///')



# -*- coding: utf-8 -*-
import os
import re
import wda
from airtestProject.airtest.utils.compat import decode_path


THISPATH = decode_path(os.path.dirname(os.path.realpath(__file__)))
DEFAULT_IPROXY_PATH = {
    "Windows": os.path.join(THISPATH, "iproxy", "windows", "iproxy.exe"),
    "Darwin": os.path.join(THISPATH, "iproxy", "mac", "iproxy"),
}
DEBUG = True
IP_PATTERN = re.compile(r'(\d+\.){3}\d+')

# When some devices (6P/7P/8P) are in landscape mode, the desktop will also change to landscape mode,
# but the click coordinates are vertical screen coordinates and require special processing
# 部分设备（6P/7P/8P）在横屏时，桌面也会变成横屏，但是点击坐标是竖屏坐标，需要特殊处理
# 由于wda不能获取到手机型号，暂时用屏幕尺寸来识别是否是plus手机
# https://developer.apple.com/design/human-interface-guidelines/ios/visual-design/adaptivity-and-layout/
LANDSCAPE_PAD_RESOLUTION = [(1242, 2208)]


class CAP_METHOD(object):
    MINICAP = "MINICAP"
    WDACAP = "WDACAP"
    MJPEG = "MJPEG"


# wda default mjpeg server port number
DEFAULT_MJPEG_PORT = 9100

# now touch and ime only support wda
class TOUCH_METHOD(object):
    WDATOUCH = "WDATOUCH"


class IME_METHOD(object):
    WDAIME = "WDAIME"


ROTATION_MODE = {
    0: wda.PORTRAIT,
    270: wda.LANDSCAPE,
    90: wda.LANDSCAPE_RIGHT,
    180: wda.PORTRAIT_UPSIDEDOWN,
}


KEY_EVENTS = {
    "home": "home",
    "volumeup": "volumeUp",
    "volumedown": "volumeDown"
}



# -*- coding: utf-8 -*-
import warnings
from airtestProject.airtest.core.android.cap_methods.base_cap import BaseCap
from airtestProject.airtest.core.android.constant import SDK_VERISON_ANDROID7
from airtestProject.airtest import aircv


class AdbCap(BaseCap):
    def get_frame_from_stream(self):
        warnings.warn("Currently using ADB screenshots, the efficiency may be very low.")
        return self.adb.snapshot()

    def snapshot(self, ensure_orientation=True):
        screen = super(AdbCap, self).snapshot()
        if ensure_orientation and self.adb.sdk_version <= SDK_VERISON_ANDROID7:
            screen = aircv.rotate(screen, self.adb.display_info["orientation"] * 90, clockwise=False)
        return screen



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
import sys

import requests.exceptions
import wda
import time
import base64
import inspect
import logging
import traceback
from logzero import setup_logger
from functools import wraps
from urllib.parse import urlparse
from tidevice._usbmux import Usbmux
from tidevice._device import BaseDevice
from tidevice._proto import MODELS
from tidevice.exceptions import MuxError

from airtestProject.airtest import aircv
from airtestProject.airtest.core.device import Device
from airtestProject.airtest.core.ios.constant import CAP_METHOD, TOUCH_METHOD, IME_METHOD, ROTATION_MODE, KEY_EVENTS, \
    LANDSCAPE_PAD_RESOLUTION, IP_PATTERN
from airtestProject.airtest.core.ios.rotation import XYTransformer, RotationWatcher
from airtestProject.airtest.core.ios.instruct_cmd import InstructHelper
from airtestProject.airtest.utils.logger import get_logger
from airtestProject.airtest.core.ios.mjpeg_cap import MJpegcap
from airtestProject.airtest.core.settings import Settings as ST
from airtestProject.airtest.aircv.screen_recorder import ScreenRecorder, resize_by_max, get_max_size
from airtestProject.airtest.core.error import LocalDeviceError, AirtestError


LOGGING = get_logger(__name__)

DEFAULT_ADDR = "http://localhost:8100/"


def decorator_retry_session(func):
    """
    When the operation fails due to session failure, try to re-acquire the session,
    retry at most 3 times.

    当因为session失效而操作失败时，尝试重新获取session，最多重试3次。
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except (RuntimeError, wda.WDAError):
            for i in range(3):
                try:
                    self._fetch_new_session()
                    return func(self, *args, **kwargs)
                except:
                    time.sleep(0.5)
                    continue
            raise
    return wrapper


def decorator_pairing_dialog(func):
    """
    When the device is not paired, trigger the trust dialogue and try again.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except MuxError:
            LOGGING.error("Device is not yet paired. Triggered the trust dialogue. Please accept and try again." + "(iTunes is required on Windows.) " if sys.platform.startswith("win") else "")
            raise
    return wrapper


def add_decorator_to_methods(decorator):
    """
    This function takes a decorator as input and returns a decorator wrapper function. The decorator wrapper function takes a class as input and decorates all the methods of the class by applying the input decorator to each method.

    Parameters:
        - decorator: A decorator function that will be applied to the methods of the input class.

    Returns:
        - decorator_wrapper: A function that takes a class as input and decorates all the methods of the class by applying the input decorator to each method.
    """
    def decorator_wrapper(cls):
        # 获取要装饰的类的所有方法
        methods = [attr for attr in dir(cls) if callable(getattr(cls, attr)) and not attr.startswith("_")]

        # 为每个方法添加装饰器
        for method in methods:
            setattr(cls, method, decorator(getattr(cls, method)))

        return cls
    return decorator_wrapper


@add_decorator_to_methods(decorator_pairing_dialog)
class TIDevice:
    """Below staticmethods are provided by Tidevice.
    """

    @staticmethod
    def devices():
        """
        Get all available devices connected by USB, return a list of UDIDs.

        Returns:
            list: A list of UDIDs. 
            e.g. ['539c5fffb18f2be0bf7f771d68f7c327fb68d2d9']
        """
        return Usbmux().device_udid_list()
    
    @staticmethod
    def list_app(udid, app_type="user"):
        """
        Returns a list of installed applications on the device.

        Args:
            udid (str): The unique identifier of the device.
            app_type (str, optional): The type of applications to list. Defaults to "user".
                Possible values are "user", "system", or "all".

        Returns:
            list: A list of tuples containing the bundle ID, display name,
                and version of the installed applications.
            e.g. [('com.apple.mobilesafari', 'Safari', '8.0'), ...]
        """
        app_type = {
            "user": "User",
            "system": "System",
            "all": None,
        }.get(app_type.lower(), None)
        app_list = []
        for info in BaseDevice(udid, Usbmux()).installation.iter_installed(app_type=app_type):
            bundle_id = info['CFBundleIdentifier']
            try:
                display_name = info['CFBundleDisplayName']
                version = info.get('CFBundleShortVersionString', '')
                app_list.append((bundle_id, display_name, version))
            except BrokenPipeError:
                break
        return app_list

    @staticmethod
    def list_wda(udid):
        """Get all WDA on device that meet certain naming rules.

        Returns:
            List of WDA bundleID.
        """
        app_list = TIDevice.list_app(udid)
        wda_list = []
        for app in app_list:
            bundle_id, display_name, _ = app
            if (bundle_id.startswith('com.') and bundle_id.endswith(".xctrunner")) or display_name == "WebDriverAgentRunner-Runner":
                wda_list.append(bundle_id)
        return wda_list
    
    @staticmethod
    def device_info(udid):
        """
        Retrieves device information based on the provided UDID.

        Args:
            udid (str): The unique device identifier.

        Returns:
            dict: A dictionary containing selected device information. The keys include:
                - productVersion (str): The version of the product.
                - productType (str): The type of the product.
                - modelNumber (str): The model number of the device.
                - serialNumber (str): The serial number of the device.
                - phoneNumber (str): The phone number associated with the device.
                - timeZone (str): The time zone of the device.
                - uniqueDeviceID (str): The unique identifier of the device.
                - marketName (str): The market name of the device.

        """
        device_info = BaseDevice(udid, Usbmux()).device_info()
        tmp_dict = {}
        # chose some useful device info from tidevice
        """
        'DeviceName', 'ProductVersion', 'ProductType',
        'ModelNumber', 'SerialNumber', 'PhoneNumber',
        'CPUArchitecture', 'ProductName', 'ProtocolVersion',
        'RegionInfo', 'TimeIntervalSince1970', 'TimeZone',
        'UniqueDeviceID', 'WiFiAddress', 'BluetoothAddress',
        'BasebandVersion'
        """
        for attr in ('ProductVersion', 'ProductType',
            'ModelNumber', 'SerialNumber', 'PhoneNumber', 
            'TimeZone', 'UniqueDeviceID'):
            key = attr[0].lower() + attr[1:]
            if attr in device_info:
                tmp_dict[key] = device_info[attr]
        try:
            tmp_dict["marketName"] = MODELS.get(device_info['ProductType'])
        except:
            tmp_dict["marketName"] = ""
        return tmp_dict
    
    @staticmethod
    def install_app(udid, file_or_url):
        BaseDevice(udid, Usbmux()).app_install(file_or_url)

    @staticmethod
    def uninstall_app(udid, bundle_id):
        BaseDevice(udid, Usbmux()).app_uninstall(bundle_id=bundle_id)

    @staticmethod
    def start_app(udid, bundle_id):
        BaseDevice(udid, Usbmux()).app_start(bundle_id=bundle_id)

    @staticmethod
    def stop_app(udid, bundle_id):
        # Note: seems not work.
        BaseDevice(udid, Usbmux()).app_stop(pid_or_name=bundle_id)

    @staticmethod
    def ps(udid):
        """
        Retrieves the process list of the specified device.

        Parameters:
            udid (str): The unique device identifier.

        Returns:
            list: A list of dictionaries containing information about each process. Each dictionary contains the following keys:
                - pid (int): The process ID.
                - name (str): The name of the process.
                - bundle_id (str): The bundle identifier of the process.
                - display_name (str): The display name of the process.
            e.g. [{'pid': 1, 'name': 'MobileSafari', 'bundle_id': 'com.apple.mobilesafari', 'display_name': 'Safari'}, ...]

        """
        with BaseDevice(udid, Usbmux()).connect_instruments() as ts:
            app_infos = list(BaseDevice(udid, Usbmux()).installation.iter_installed(app_type=None))
            ps = list(ts.app_process_list(app_infos))
        ps_list = []
        keys = ['pid', 'name', 'bundle_id', 'display_name']
        for p in ps:
            if not p['isApplication']:
                continue
            ps_list.append({key: p[key] for key in keys})
        return ps_list
    
    @staticmethod
    def ps_wda(udid):
        """Get all running WDA on device that meet certain naming rules.

        Returns:
            List of running WDA bundleID.
        """
        with BaseDevice(udid, Usbmux()).connect_instruments() as ts:
            app_infos = list(BaseDevice(udid, Usbmux()).installation.iter_installed(app_type=None))
            ps = list(ts.app_process_list(app_infos))
        ps_wda_list = []
        for p in ps:
            if not p['isApplication']:
                continue
            if ".xctrunner" in p['bundle_id'] or p['display_name'] == "WebDriverAgentRunner-Runner":
                ps_wda_list.append(p['bundle_id'])
            else:
                continue
        return ps_wda_list
    
    @staticmethod
    def xctest(udid, wda_bundle_id):
        return BaseDevice(udid, Usbmux()).xctest(fuzzy_bundle_id=wda_bundle_id, logger=setup_logger(level=logging.INFO))


@add_decorator_to_methods(decorator_retry_session)
class IOS(Device):
    """IOS client.

        - before this you have to run `WebDriverAgent <https://github.com/AirtestProject/iOS-Tagent>`_

        - ``xcodebuild -project path/to/WebDriverAgent.xcodeproj -scheme WebDriverAgentRunner -destination "id=$(idevice_id -l)" test``

        - ``iproxy $port 8100 $udid``
    """

    def __init__(self, addr=DEFAULT_ADDR, cap_method=CAP_METHOD.MJPEG, mjpeg_port=None, udid=None, name=None, serialno=None, wda_bundle_id=None):
        super().__init__()

        # If none or empty, use default addr.
        self.addr = addr or DEFAULT_ADDR

        # Fit wda format, make url start with http://,
        # e.g., http://localhost:8100/ or http+usbmux://00008020-001270842E88002E
        # with mjpeg_port: http://localhost:8100/?mjpeg_port=9100
        # with udid(udid/uuid/serialno are all ok): http://localhost:8100/?mjpeg_port=9100&&udid=00008020-001270842E88002E.
        if not self.addr.startswith("http"):
            self.addr = "http://" + addr

        # Here now use these supported cap touch and ime method.
        self.cap_method = cap_method
        self.touch_method = TOUCH_METHOD.WDATOUCH
        self.ime_method = IME_METHOD.WDAIME

        # Wda driver, use to home, start app, click/swipe/close app/get wda size.
        # Init wda session, updata when start app.
        wda.DEBUG = False
        # The three connection modes are determined by the addr.
        # e.g., connect remote device http://10.227.70.247:20042
        # e.g., connect local device http://127.0.0.1:8100 or http://localhost:8100 or http+usbmux://00008020-001270842E88002E
        self.udid = udid or name or serialno
        self._wda_bundle_id = wda_bundle_id
        parsed = urlparse(self.addr).netloc.split(":")[0] if ":" in urlparse(self.addr).netloc else urlparse(self.addr).netloc
        if parsed not in ["localhost", "127.0.0.1"] and "." in parsed:
            # Connect remote device via url.
            self.is_local_device = False
            self.driver = wda.Client(self.addr) 
        else:   
            # Connect local device via url.
            self.is_local_device = True
            if parsed in ["localhost", "127.0.0.1"]:
                if not udid:
                    udid = self._get_default_device()
                self.udid = udid
            else:
                self.udid = parsed
            self.driver = wda.USBClient(udid=self.udid, port=8100, wda_bundle_id=self.wda_bundle_id)
        # Record device's width and height.
        self._size = {'width': None, 'height': None}
        self._current_orientation = None
        self._touch_factor = None
        self._last_orientation = None
        self._is_pad = None
        self._using_ios_tagent = None
        self._device_info = {}
        self.instruct_helper = InstructHelper(self.device_info['uuid'])
        self.mjpegcap = MJpegcap(self.instruct_helper, ori_function=lambda: self.display_info,
                                 ip=self.ip, port=mjpeg_port)
        # Start up RotationWatcher with default session.
        self.rotation_watcher = RotationWatcher(self)
        self._register_rotation_watcher()

        self.alert_watch_and_click = self.driver.alert.watch_and_click

        # Recorder.
        self.recorder = None

        # Since uuid and udid are very similar, both names are allowed.
        self._udid = udid or name or serialno

    def _get_default_device(self):
        """Get local default device when no udid.

        Returns:
            Local device udid.
        """
        device_udid_list = TIDevice.devices()
        if device_udid_list:
            return device_udid_list[0]
        raise IndexError("iOS devices not found, please connect device first.")
    
    def _get_default_wda_bundle_id(self):
        """Get local default device's WDA bundleID when no WDA bundleID.

        Returns:
            Local device's WDA bundleID.
        """ 
        try:
            wda_list = TIDevice.list_wda(self.udid)
            return wda_list[0]
        except IndexError:
            raise IndexError("WDA bundleID not found, please install WDA on device.")
        
    def _get_default_running_wda_bundle_id(self):
        """Get the bundleID of the WDA that is currently running on local device.

        Returns:
            Local device's running WDA bundleID.
        """ 
        try:
            running_wda_list = TIDevice.ps_wda(self.udid)
            return running_wda_list[0]
        except IndexError:
            raise IndexError("Running WDA bundleID not found, please makesure WDA was started.")

    @property
    def wda_bundle_id(self):
        if not self._wda_bundle_id and self.is_local_device:
            self._wda_bundle_id = self._get_default_wda_bundle_id()
        return self._wda_bundle_id
        
    @property
    def ip(self):
        """Returns the IP address of the host connected to the iOS phone.

        It is not the IP address of the iOS phone.
        If you want to get the IP address of the phone, you can access the interface `get_ip_address`.
        For example: when the device is connected via http://localhost:8100, return localhost.
        If it is a remote device http://192.168.xx.xx:8100, it returns the IP address of 192.168.xx.xx.

        Returns:
            IP.
        """
        match = re.search(IP_PATTERN, self.addr)
        if match:
            ip = match.group(0)
        else:
            ip = 'localhost'
        return ip

    @property
    def uuid(self):
        return self._udid or self.addr

    @property
    def using_ios_tagent(self):
        """
        当前基础版本：appium/WebDriverAgent 4.1.4
        基于上述版本，2022.3.30之后发布的iOS-Tagent版本，在/status接口中新增了一个Version参数，如果能检查到本参数，说明使用了新版本ios-Tagent
        该版本基于Appium版的WDA做了一些改动，可以更快地进行点击和滑动，并优化了部分UI树的获取逻辑
        但是所有的坐标都为竖屏坐标，需要airtest自行根据方向做转换
        同时，大于4.1.4版本的appium/WebDriverAgent不再需要针对ipad进行特殊的横屏坐标处理了

        Returns:
            True if using ios-tagent, else False.
        """
        if self._using_ios_tagent is None:
            status = self.driver.status()
            if 'Version' in status:
                self._using_ios_tagent = True
            else:
                self._using_ios_tagent = False
        return self._using_ios_tagent

    def _fetch_new_session(self):
        """Re-acquire a new session.
        """
        # 根据facebook-wda的逻辑，直接设为None就会自动获取一个新的默认session
        self.driver.session_id = None

    @property
    def is_pad(self):
        """
        Determine whether it is an ipad(or 6P/7P/8P), if it is, in the case of horizontal screen + desktop,
        the coordinates need to be switched to vertical screen coordinates to click correctly (WDA bug).

        判断是否是ipad(或 6P/7P/8P)，如果是，在横屏+桌面的情况下，坐标需要切换成竖屏坐标才能正确点击（WDA的bug）

        Returns:
            True if it is an ipad, else False.
        """
        if self._is_pad is None:
            info = self.device_info
            if info["model"] == "iPad" or \
                    (self.display_info["width"], self.display_info["height"]) in LANDSCAPE_PAD_RESOLUTION:
                # ipad与6P/7P/8P等设备，桌面横屏时的表现一样，都会变横屏
                self._is_pad = True
            else:
                self._is_pad = False
        return self._is_pad

    @property
    def device_info(self):
        """ Get the device info.

        Notes:
            Might not work on all devices.

        Returns:
            Dict for device info,
            e.g. AttrDict({
                'timeZone': 'GMT+0800',
                'currentLocale': 'zh_CN',
                'model': 'iPhone',
                'uuid': '90CD6AB7-11C7-4E52-B2D3-61FA31D791EC',
                'userInterfaceIdiom': 0,
                'userInterfaceStyle': 'light',
                'name': 'iPhone',
                'isSimulator': False})
        """
        if not self._device_info:
            device_info = self.driver.info
            try:
                # Add some device info.
                if not self.is_local_device:
                    raise LocalDeviceError()
                tmp_dict = TIDevice.device_info(self.udid)
                device_info.update(tmp_dict)
            except:
                pass
            finally:
                self._device_info = device_info
        return self._device_info

    def _register_rotation_watcher(self):
        """
        Register callbacks for Android and minicap when rotation of screen has changed,
        callback is called in another thread, so be careful about thread-safety.
        """
        self.rotation_watcher.reg_callback(lambda x: setattr(self, "_current_orientation", x))

    def window_size(self):
        """
        Returns: 
            Window size (width, height).
        """
        try:
            return self.driver.window_size()
        except wda.exceptions.WDAStaleElementReferenceError:
            print("iOS connection failed, please try pressing the home button to return to the desktop and try again.")
            print("iOS连接失败，请尝试按home键回到桌面后再重试")
            raise

    @property
    def orientation(self):
        """
        Returns:
            Device orientation status in LANDSACPE POR.
        """
        if not self._current_orientation:
            self._current_orientation = self.get_orientation()
        return self._current_orientation

    @orientation.setter
    def orientation(self, value):
        """
        Args:
            value(string): LANDSCAPE | PORTRAIT | UIA_DEVICE_ORIENTATION_LANDSCAPERIGHT |
                    UIA_DEVICE_ORIENTATION_PORTRAIT_UPSIDEDOWN
        """
        # 可以对屏幕进行旋转，但是可能会导致问题
        self.driver.orientation = value

    def get_orientation(self):
        # self.driver.orientation只能拿到LANDSCAPE，不能拿到左转/右转的确切方向
        # 因此手动调用/rotation获取屏幕实际方向
        rotation = self.driver._session_http.get('/rotation')
        # rotation dict eg. {'value': {'x': 0, 'y': 0, 'z': 90}, 'sessionId': 'xx', 'status': 0}
        if rotation:
            return ROTATION_MODE.get(rotation['value']['z'], wda.PORTRAIT)

    @property
    def display_info(self):
        if not self._size['width'] or not self._size['height']:
            self._display_info()

        self._size['orientation'] = self.orientation
        return self._size

    def _display_info(self):
        # Function window_size() return UIKit size, while screenshot() image size is Native Resolution.
        window_size = self.window_size()
        # When use screenshot, the image size is pixels size. e.g.(1080 x 1920)
        snapshot = self.snapshot()
        if self.orientation in [wda.LANDSCAPE, wda.LANDSCAPE_RIGHT]:
            self._size['window_width'], self._size['window_height'] = window_size.height, window_size.width
            width, height = snapshot.shape[:2]
        else:
            self._size['window_width'], self._size['window_height'] = window_size.width, window_size.height
            height, width = snapshot.shape[:2]
        self._size["width"] = width
        self._size["height"] = height

        # Use session.scale can get UIKit scale factor.
        # So self._touch_factor = 1 / self.driver.scale, but the result is incorrect on some devices(6P/7P/8P).
        self._touch_factor = float(self._size['window_height']) / float(height)
        self.rotation_watcher.get_ready()

        # 当前版本: wda 4.1.4 获取到的/window/size，在ipad+桌面横屏下拿到的是 height * height，需要修正。
        if self.is_pad and self.home_interface():
            self._size['window_width'] = int(width * self._touch_factor)

    @property
    def touch_factor(self):
        if not self._touch_factor:
            self._display_info()
        return self._touch_factor

    @touch_factor.setter
    def touch_factor(self, factor):
        """
        Func touch_factor is used to convert click coordinates: mobile phone real coordinates = touch_factor * screen coordinates.
        In general, no manual settings are required.

        touch_factor用于换算点击坐标：手机真实坐标 = touch_factor * 屏幕坐标
        默认计算方式是： self.display_info['window_height'] / self.display_info['height']
        但在部分特殊型号手机上可能不准确，例如iOS14.4的7P，默认值为 1/3，但部分7P点击位置不准确，可自行设置为：self.touch_factor = 1 / 3.3
        （一般情况下，无需手动设置！）

        Examples:
            >>> device = connect_device("iOS:///")
            >>> device.touch((100, 100))  # wrong position
            >>> print(device.touch_factor)
            0.333333
            >>> device.touch_factor = 1 / 3.3  # default is 1/3
            >>> device.touch((100, 100))

        Args:
            factor: real_pos / pixel_pos, e.g: 1/self.driver.scale
        """
        self._touch_factor = float(factor)

    def get_render_resolution(self):
        """Return render resolution after rotation.

        Returns:
            offset_x, offset_y, offset_width and offset_height of the display.
        """
        w, h = self.get_current_resolution()
        return 0, 0, w, h

    def get_current_resolution(self):
        w, h = self.display_info["width"], self.display_info["height"]
        if self.display_info["orientation"] in [wda.LANDSCAPE, wda.LANDSCAPE_RIGHT]:
            w, h = h, w
        return w, h

    def home(self):
        # Press("home") faster than home().
        return self.driver.press("home")

    def _neo_wda_screenshot(self):
        """
        This is almost same as wda implementation, but without png header check,
        as response data is now jpg format in mid quality.
        """
        value = self.driver.http.get('screenshot').value
        raw_value = base64.b64decode(value)
        return raw_value

    def snapshot(self, filename=None, quality=10, max_size=None):
        """Take snapshot.

        Args:
            filename: save screenshot to filename
            quality: The image quality, integer in range [1, 99]
            max_size: the maximum size of the picture, e.g 1200

        Returns:
            Screen snapshot's cv2 object.
        """
        data = self._neo_wda_screenshot()
        # Output cv2 object.
        try:
            screen = aircv.utils.string_2_img(data)
        except:
            # May be black/locked screen or other reason, print exc for debugging.
            traceback.print_exc()
            return None

        # Save as file if needed.
        if filename:
            aircv.imwrite(filename, screen, quality, max_size=max_size)

        return screen

    def get_frame_from_stream(self):
        if self.cap_method == CAP_METHOD.MJPEG:
            try:
                return self.mjpegcap.get_frame_from_stream()
            except ConnectionRefusedError:
                self.cap_method = CAP_METHOD.WDACAP
        return self._neo_wda_screenshot()

    def touch(self, pos, duration=0.01):
        """
        Args:
            pos: coordinates (x, y), can be float(percent) or int
            duration (optional): tap_hold duration

        Returns: None(iOS may all use vertical screen coordinates, so coordinates will not be returned.)

        Examples:
            >>> touch((100, 100))
            >>> touch((0.5, 0.5), duration=1)

        """
        x, y = pos
        if not (x < 1 and y < 1):
            x, y = int(x * self.touch_factor), int(y * self.touch_factor)
        # 根据用户安装的wda版本判断是否调用快速点击接口
        if self.using_ios_tagent:
            x, y = self._transform_xy(pos)
            self._quick_click(x, y, duration)
        else:
            self.driver.click(x, y, duration)

    def _quick_click(self, x, y, duration):
        """
        The method extended from the facebook-wda third-party library.
        Use modified appium wda to perform quick click.
        
        Args:
            x, y (int, float): float(percent), int(coordicate)
            duration (optional): tap_hold duration
        """
        x, y = self.driver._percent2pos(x, y)
        data = {'x': x, 'y': y, 'duration': duration}
        # 为了兼容改动直接覆盖原生接口的自制版wda。
        try:
            return self.driver._session_http.post('/wda/tap', data=data)
        except wda.WDARequestError as e:
            if e.status == 110:
                self.driver.click(x, y, duration)
                
    def double_click(self, pos):
        x, y = self._transform_xy(pos)
        self.driver.double_tap(x, y)

    def swipe(self, fpos, tpos, duration=0, delay=None, *args, **kwargs):
        """
        Args:
            fpos: start point
            tpos: end point
            duration (float): start coordinate press duration (seconds), default is None
            delay (float): start coordinate to end coordinate duration (seconds)

        Returns:
            None

        Examples:
            >>> swipe((1050, 1900), (150, 1900))
            >>> swipe((0.2, 0.5), (0.8, 0.5))
        """
        fx, fy = fpos
        tx, ty = tpos
        if not (fx < 1 and fy < 1):
            fx, fy = int(fx * self.touch_factor), int(fy * self.touch_factor)
        if not (tx < 1 and ty < 1):
            tx, ty = int(tx * self.touch_factor), int(ty * self.touch_factor)
        # 如果是通过ide来滑动，且安装的是自制版的wda就调用快速滑动接口，其他时候不关心滑动速度就使用原生接口保证滑动精确性。
        def ios_tagent_swipe(fpos, tpos, delay=None):
            # 调用自定义的wda swipe接口需要进行坐标转换。
            fx, fy = self._transform_xy(fpos)
            tx, ty = self._transform_xy(tpos)
            self._quick_swipe(fx, fy, tx, ty, delay or 0.2)

        if delay:
            if self.using_ios_tagent:
                ios_tagent_swipe(fpos, tpos, delay)
            else:
                self.driver.swipe(fx, fy, tx, ty)
        else:
            try:
                self.driver.swipe(fx, fy, tx, ty, duration)
            except wda.WDARequestError as e:
                if e.status == 110:
                    # 部分低版本iOS在滑动时，可能会有报错，调用一次ios-Tagent的快速滑动接口来尝试解决
                    if self.using_ios_tagent:
                        ios_tagent_swipe(fpos, tpos, delay)
                    else:
                        raise

    def _quick_swipe(self, x1, y1, x2, y2, delay):
        """
        The method extended from the facebook-wda third-party library.
        Use modified appium wda to perform quick swipe.
        
        Args:
            x1, y1, x2, y2 (int, float): float(percent), int(coordicate)
            delay (float): start coordinate to end coordinate duration (seconds)
        """
        if any(isinstance(v, float) for v in [x1, y1, x2, y2]):
            size = self.window_size()
            x1, y1 = self.driver._percent2pos(x1, y1, size)
            x2, y2 = self.driver._percent2pos(x2, y2, size)

        data = dict(fromX=x1, fromY=y1, toX=x2, toY=y2, delay=delay)
        # 为了兼容改动直接覆盖原生接口的自制版wda。
        try:
            return self.driver._session_http.post('/wda/swipe', data=data)
        except wda.WDARequestError as e:
            if e.status == 110:
                self.driver.swipe(x1, y1, x2, y2)

    def keyevent(self, keyname, **kwargs):
        """Perform keyevent on the device.

        Args:
            keyname: home/volumeUp/volumeDown
            **kwargs:
        """
        try:
            keyname = KEY_EVENTS[keyname.lower()]
        except KeyError:
            raise ValueError("Invalid name: %s, should be one of ('home', 'volumeUp', 'volumeDown')" % keyname)
        else:
            self.press(keyname)

    def press(self, keys):
        """Some keys in ["home", "volumeUp", "volumeDown"] can be pressed.
        """
        self.driver.press(keys)

    def text(self, text, enter=True):
        """Input text on the device.

        Args:
            text:  text to input
            enter: True if you need to enter a newline at the end

        Examples:
            >>> text("test")
            >>> text("中文")
        """
        if enter:
            text += '\n'
        self.driver.send_keys(text)
        
    def install_app(self, file_or_url, **kwargs):
        """
        curl -X POST $JSON_HEADER \
        -d "{\"desiredCapabilities\":{\"bundleId\":\"com.apple.mobilesafari\", \"app\":\"[host_path]/magicapp.app\"}}" \
        $DEVICE_URL/session
        https://github.com/facebook/WebDriverAgent/wiki/Queries

        Install app from the device.

        Args:
            file_or_url: file or url to install

        Returns:
            bundle ID

        Raises:
            LocalDeviceError: If the device is remote.
        """
        if not self.is_local_device:
            raise LocalDeviceError()
        return TIDevice.install_app(self.udid, file_or_url)
    
    def uninstall_app(self, bundle_id):
        """Uninstall app from the device.

        Notes: 
            It seems always return True.

        Args:
            bundle_id: the app bundle id, e.g ``com.apple.mobilesafari``

        Raises:
            LocalDeviceError: If the device is remote.
        """
        if not self.is_local_device:
            raise LocalDeviceError()
        return TIDevice.uninstall_app(self.udid, bundle_id)

    def start_app(self, bundle_id, *args, **kwargs):
        """
        Args:
            bundle_id: the app bundle id, e.g ``com.apple.mobilesafari``

        Returns:
            Process ID.

        Examples:
            >>> start_app('com.apple.mobilesafari')
        """
        if not self.is_local_device:
            # Note: If the bundle_id does not exist, it may get stuck.
            try:
                return self.driver.app_launch(bundle_id)
            except requests.exceptions.ReadTimeout:
                raise AirtestError(f"App launch timeout, please check if the app is installed: {bundle_id}")
        else:
            return TIDevice.start_app(self.udid, bundle_id)
    
    def stop_app(self, bundle_id):
        """
        Note: Both ways of killing the app may fail, nothing responds or just closes the 
        app to the background instead of actually killing it and no error will be reported.
        """
        try:
            if not self.is_local_device:
                raise LocalDeviceError()
            TIDevice.stop_app(self.udid, bundle_id)
        except:
            pass
        finally:
            self.driver.app_stop(bundle_id=bundle_id)

    def list_app(self, type="user"):
        """
        Returns a list of installed applications on the device.

        Args:
            type (str, optional): The type of applications to list. Defaults to "user".
                Possible values are "user", "system", or "all".

        Returns:
            list: A list of tuples containing the bundle ID, display name,
                and version of the installed applications.
            e.g. [('com.apple.mobilesafari', 'Safari', '8.0'), ...]

        Raises:
            LocalDeviceError: If the device is remote.
        """
        if not self.is_local_device:
            raise LocalDeviceError()
        return TIDevice.list_app(self.udid, app_type=type)

    def app_state(self, bundle_id):
        """ Get app state and ruturn.

        Args:
            bundle_id: Bundle ID of app.

        Returns:
            {
                "value": 4,
                "sessionId": "0363BDC5-4335-47ED-A54E-F7CCB65C6A65"
            }

            Value 1(not running) 2(running in background) 3(running in foreground)? 4(running).
        """
        return self.driver.app_state(bundle_id=bundle_id)

    def app_current(self):
        """Get the app current.

        Notes:
            Might not work on all devices.

        Returns:
            Current app state dict, eg:
            {"pid": 1281,
             "name": "",
             "bundleId": "com.netease.cloudmusic"}
        """
        return self.driver.app_current()
    
    def get_clipboard(self, wda_bundle_id=None, *args, **kwargs):
        """Get clipboard text.

        Before calling the WDA interface, you need to ensure that WDA was foreground.  
        If there are multiple WDA on your device, please specify the active WDA by parameter wda_bundle_id.
        
        Args:
            wda_bundle_id: The bundle id of the running WDA, if None, will use default WDA bundle id.

        Returns:
            Clipboard text.

        Raises:
            LocalDeviceError: If the device is remote and the wda_bundle_id parameter is not provided.

        Notes:
            If you want to use this function, you have to set WDA foreground which would switch the 
            current screen of the phone. Then we will try to switch back to the screen before.
        """
        if wda_bundle_id is None:
            if not self.is_local_device:
                raise LocalDeviceError("Remote device need to set running wda bundle id parameter, \
                                       e.g. get_clipboard('wda_bundle_id').")
            wda_bundle_id = self.wda_bundle_id
        # Set wda foreground, it's necessary.
        try:
            current_app_bundle_id = self.app_current().get("bundleId", None)
        except:
            current_app_bundle_id = None
        try:
            self.driver.app_launch(wda_bundle_id)
        except:
            pass
        # 某些机型版本下，有一定概率获取失败，尝试重试一次
        clipboard_text = self.driver._session_http.post("/wda/getPasteboard").value
        if not clipboard_text:
            clipboard_text = self.driver._session_http.post("/wda/getPasteboard").value
        decoded_text = base64.b64decode(clipboard_text).decode('utf-8')

        # Switch back to the screen before.
        if current_app_bundle_id:
            self.driver.app_launch(current_app_bundle_id)
        else:
            LOGGING.warning("we can't switch back to the app before, because can't get bundle id.")
        return decoded_text
    
    def set_clipboard(self, content, wda_bundle_id=None, *args, **kwargs):
        """
        Set the clipboard content on the device.

        Args:
            content (str): The content to be set on the clipboard.
            wda_bundle_id (str, optional): The bundle ID of the WDA app. Defaults to None.

        Raises:
            LocalDeviceError: If the device is remote and the wda_bundle_id parameter is not provided.

        Returns:
            None
        """
        if wda_bundle_id is None:
            if not self.is_local_device:
                raise LocalDeviceError("Remote device need to set running wda bundle id parameter, \
                                        e.g. set_clipboard('content', 'wda_bundle_id').")
            wda_bundle_id = self.wda_bundle_id
        # Set wda foreground, it's necessary.
        try:
            current_app_bundle_id = self.app_current().get("bundleId", None)
        except:
            current_app_bundle_id = None
        try:
            self.driver.app_launch(wda_bundle_id)
        except:
            pass
        self.driver.set_clipboard(content)
        clipboard_text = self.driver._session_http.post("/wda/getPasteboard").value
        decoded_text = base64.b64decode(clipboard_text).decode('utf-8')
        if decoded_text != content:
            # 部分机型偶现设置剪切板失败，重试一次
            self.driver.set_clipboard(content)
        # Switch back to the screen before.
        if current_app_bundle_id:
            self.driver.app_launch(current_app_bundle_id)
        else:
            LOGGING.warning("we can't switch back to the app before, because can't get bundle id.")

    def paste(self, wda_bundle_id=None, *args, **kwargs):
        """
        Paste the current clipboard content on the device.

        Args:
            wda_bundle_id (str, optional): The bundle ID of the WDA app. Defaults to None.

        Raises:
            LocalDeviceError: If the device is remote and the wda_bundle_id parameter is not provided.

        Returns:
            None
        """
        self.text(self.get_clipboard(wda_bundle_id=wda_bundle_id))

    def get_ip_address(self):
        """Get ip address from WDA.

        Returns:
            If no IP address has been found, otherwise return the IP address.
        """
        ios_status = self.driver.status()['ios']
        # If use modified WDA, try to get real wifi IP first.
        return ios_status.get('wifiIP', ios_status['ip'])

    def device_status(self):
        """Show status return by WDA.

        Returns:
            Dicts of infos.
        """
        return self.driver.status()

    def _touch_point_by_orientation(self, tuple_xy):
        """
        Convert image coordinates to physical display coordinates, the arbitrary point (origin) is upper left corner
        of the device physical display.

        Args:
            tuple_xy: image coordinates (x, y)

        Returns:
            physical coordinates (x, y)
        """
        x, y = tuple_xy

        # 1. 如果使用了2022.03.30之后发布的iOS-Tagent版本，则必须要进行竖屏坐标转换。
        # 2. 如果使用了appium/WebDriverAgent>=4.1.4版本，直接使用原坐标即可，无需转换。
        # 3. 如果使用了appium/WebDriverAgent<4.1.4版本，或低版本的iOS-Tagent，并且ipad下横屏点击异常，请改用airtest<=1.2.4。
        if self.using_ios_tagent:
            width = self.display_info["width"]
            height = self.display_info["height"]
            if self.orientation in [wda.LANDSCAPE, wda.LANDSCAPE_RIGHT]:
                width, height = height, width
            if x < 1 and y < 1:
                x = x * width
                y = y * height
            x, y = XYTransformer.up_2_ori(
                (x, y),
                (width, height),
                self.orientation
            )
        return x, y

    def _transform_xy(self, pos):
        x, y = self._touch_point_by_orientation(pos)

        # Scale touch postion.
        if not (x < 1 and y < 1):
            x, y = int(x * self.touch_factor), int(y * self.touch_factor)

        return x, y

    def _check_orientation_change(self):
        pass

    def is_locked(self):
        """
        Return True or False whether the device is locked or not.

        Notes:
            Might not work on some devices.

        Returns:
            True or False.
        """
        return self.driver.locked()

    def unlock(self):
        """Unlock the device, unlock screen, double press home.

        Notes:
            Might not work on all devices.
        """
        return self.driver.unlock()

    def lock(self):
        """Lock the device, lock screen.

        Notes:
            Might not work on all devices.
        """
        return self.driver.lock()

    def setup_forward(self, port):
        """
        Setup port forwarding from device to host.
        Args:
            port: device port

        Returns:
            host port, device port

        Raises:
            LocalDeviceError: If the device is remote.

        """
        return self.instruct_helper.setup_proxy(int(port))

    def ps(self):
        """Get the process list of the device.

        Returns:
            Process list of the device.
        """
        if not self.is_local_device:
            raise LocalDeviceError()
        return TIDevice.ps(self.udid)

    def alert_accept(self):
        """ Alert accept-Actually do click first alert button.

        Notes:
            Might not work on all devices.
        """
        return self.driver.alert.accept()

    def alert_dismiss(self):
        """Alert dissmiss-Actually do click second alert button.

        Notes:
            Might not work on all devices.
        """
        return self.driver.alert.dismiss()

    def alert_wait(self, time_counter=2):
        """If alert apper in time_counter second it will return True,else return False (default 20.0).

        Notes:
            Might not work on all devices.
        """
        return self.driver.alert.wait(time_counter)

    def alert_buttons(self):
        """Get alert buttons text. 

        Notes:
            Might not work on all devices.

        Returns:
             # example return: ("设置", "好")

        """
        return self.driver.alert.buttons()

    def alert_exists(self):
        """ Get True for alert exists or False.

        Notes:
            Might not work on all devices

        Returns:
            True or False
        """
        return self.driver.alert.exists

    def alert_click(self, buttons):
        """When Arg type is list, click the first match, raise ValueError if no match.

        e.g. ["设置", "信任", "安装"]

        Notes:
            Might not work on all devices.
        """
        return self.driver.alert.click(buttons)

    def home_interface(self):
        """Get True for the device status is on home interface. 

        Reason:
            Some devices can Horizontal screen on the home interface.

        Notes:
            Might not work on all devices.

        Returns:
            True or False
        """
        try:
            app_current_dict = self.app_current()
            app_current_bundleId = app_current_dict.get('bundleId')
            LOGGING.info("app_current_bundleId %s", app_current_bundleId)
        except:
            return False
        else:
            if app_current_bundleId in ['com.apple.springboard']:
                return True
        return False

    def disconnect(self):
        """Disconnected mjpeg and rotation_watcher.
        """
        if self.cap_method == CAP_METHOD.MJPEG:
            self.mjpegcap.teardown_stream()
        if self.rotation_watcher:
            self.rotation_watcher.teardown()
    
    def start_recording(self, max_time=1800, output=None, fps=10,
                        snapshot_sleep=0.001, orientation=0, max_size=None, *args, **kwargs):
        """Start recording the device display.

        Args:
            max_time: maximum screen recording time, default is 1800
            output: ouput file path
            fps: frames per second will record
            snapshot_sleep: sleep time for each snapshot.
            orientation: 1: portrait, 2: landscape, 0: rotation, default is 0
            max_size: max size of the video frame, e.g.800, default is None. Smaller sizes lead to lower system load.

        Returns:
            save_path: path of video file

        Examples:

            Record 30 seconds of video and export to the current directory test.mp4:

            >>> from airtest.core.api import connect_device, sleep
            >>> dev = connect_device("IOS:///")
            >>> save_path = dev.start_recording(output="test.mp4")
            >>> sleep(30)
            >>> dev.stop_recording()
            >>> print(save_path)

            >>> # the screen is portrait
            >>> portrait_mp4 = dev.start_recording(output="portrait.mp4", orientation=1)  # or orientation="portrait"
            >>> sleep(30)
            >>> dev.stop_recording()

            >>> # the screen is landscape
            >>> landscape_mp4 = dev.start_recording(output="landscape.mp4", orientation=2)  # or orientation="landscape"

            You can specify max_size to limit the video's maximum width/length. Smaller video sizes result in lower CPU load.

            >>> dev.start_recording(output="test.mp4", max_size=800)

        """
        if fps > 10 or fps < 1:
            LOGGING.warning("fps should be between 1 and 10, becuase of the recording effiency")
            if fps > 10:
                fps = 10
            if fps < 1:
                fps = 1

        if self.recorder and self.recorder.is_running():
            LOGGING.warning("recording is already running, please don't call again")
            return None
        
        logdir = "./"
        if ST.LOG_DIR is not None:
            logdir = ST.LOG_DIR
        if output is None:
            save_path = os.path.join(logdir, "screen_%s.mp4" % (time.strftime("%Y%m%d%H%M%S", time.localtime())))
        else:
            if os.path.isabs(output):
                save_path = output
            else:
                save_path = os.path.join(logdir, output)

        max_size = get_max_size(max_size)
        def get_frame():
            data = self.get_frame_from_stream()
            frame = aircv.utils.string_2_img(data)
            
            if max_size is not None:
                frame = resize_by_max(frame, max_size)
            return frame

        self.recorder = ScreenRecorder(
            save_path, get_frame, fps=fps,
            snapshot_sleep=snapshot_sleep, orientation=orientation)
        self.recorder.stop_time = max_time
        self.recorder.start()
        LOGGING.info("start recording screen to {}".format(save_path))
        return save_path

    def stop_recording(self,):
        """ Stop recording the device display. Recoding file will be kept in the device.
        """
        LOGGING.info("stopping recording")
        self.recorder.stop()
        return None



import threading
from functools import wraps


class ThreadSafeIter:
    """
    Takes an iterator/generator and makes it thread-safe by
    serializing call to the `next` method of given iterator/generator.
    """
    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()
        if getattr(self.it, "__next__", None) is None:  # for py2
            self._next = self.it.next
        else:
            self._next = self.it.__next__  # py3

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return self._next()

    def send(self, *args):
        with self.lock:
            return self.it.send(*args)

    next = __next__  # for Python 2


def threadsafe_generator(f):
    """
    A decorator that takes a generator function and makes it thread-safe.
    """
    @wraps(f)
    def g(*a, **kw):
        return ThreadSafeIter(f(*a, **kw))
    return g



#!-*- coding = utf-8 -*-
# @Time : 2024/4/28 8:08
# @Author : 苏嘉浩
# @File : ocrTemplate.py
# @Software : PyCharm
import threading

import cv2
import easyocr
import numpy as np

import torch
from airtestProject.airtest import aircv
from airtestProject.airtest.core.cv import Template
from airtestProject.airtest.core.error import InvalidMatchingMethodError
from airtestProject.airtest.utils.transform import TargetPos
from airtestProject.airtest.core.helper import G, logwrap
from airtestProject.airtest.core.settings import Settings as ST  # noqa
from paddleocr import PaddleOCR
from airtestProject.commons.utils.setting import Setting as SS
import os

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"


class EasyOcr:
    def __init__(self, language=None):
        """
        :param language: 语言默认为中文和英文。可能会比较慢，后续可以建议为单引擎
        """
        if language is None:
            self.easy_language = ['en', 'ch_sim']
        else:
            self.easy_language = list(language)
            index = self.easy_language.index('ch')  # 找到 'ch' 的位置,easyocr的中文是分繁体简体
            self.easy_language[index] = 'ch_sim'
        self.reader = easyocr.Reader(self.easy_language, gpu=True)

    def ocr_match(self, img, input_text):
        # 读取图像

        result = self.reader.readtext(img)
        # 转为 {'result': (1606, 1116), 'rectangle': [(1271, 1046), (1271, 1187), (1941, 1187), (1941, 1046)],
        # 'confidence': 0.8403522074222565, 'time': 0.7186686992645264}
        result_list = []
        for res in result:
            result_dict = {}
            (bbox, text, prob) = res
            bbox = [(float(point[0]), float(point[1])) for point in bbox]
            (tl, tr, br, bl) = bbox
            center_x = (tl[0] + br[0]) / 2
            center_y = (tl[1] + br[1]) / 2
            result_dict['result'] = (center_x, center_y)
            result_dict['rectangle'] = bbox
            result_dict['confidence'] = prob
            result_dict['text'] = text
            result_list.append(result_dict)
        # for result_dict1 in result_list:
        #     if result_dict1['text'] == input_text:
        #         return result_dict1['result']
        # return False
        # 生成式等效于for循环
        print(result_list)
        coordinate = next((r_dict for r_dict in result_list if r_dict.get('text') == input_text), None)
        return coordinate


class PpOcr:
    """
    :param language: 默认为双引擎，会比较慢，可以针对项目启用单引擎（Odin中英文都有很难受）
    """

    def __init__(self, language=None):

        if language is None:
            self.ppOcr_languages = ['ch']
        else:
            self.ppOcr_languages = list(language)
        self.ocr_modules = [PaddleOCR(use_angle_cls=True, use_gpu=True, lang=lang) for lang in self.ppOcr_languages]

    def ocr_match(self, img, input_text):
        pp_results = {}
        for ocr_model in self.ocr_modules:
            result = ocr_model.ocr(img, cls=True)
            for re in result:
                for line in re:
                    text, coordinates, confidence = line[1][0], line[0], line[1][1]
                    coordinates_key = ''.join(map(str, coordinates))
                    # 如果文本,还需要比对坐标是否相同不然会已经在字典中，比较可信度
                    # print((text, tuple(coordinates)) in pp_results)
                    if (text, coordinates_key) in pp_results and confidence <= pp_results[(text, coordinates_key)][0]:
                        continue
                    # 如果文本不在字典中，或者新的可信度更高，更新字典
                    pp_results[text, coordinates_key] = (confidence, coordinates)
        result_list = []
        for key, value in pp_results.items():
            result_dict = {}
            (prob, bbox) = value
            bbox = [(float(point[0]), float(point[1])) for point in bbox]
            (tl, tr, br, bl) = bbox
            center_x = (tl[0] + br[0]) / 2
            center_y = (tl[1] + br[1]) / 2
            result_dict['result'] = (center_x, center_y)
            result_dict['rectangle'] = bbox
            result_dict['confidence'] = prob
            result_dict['text'] = key[0]
            result_list.append(result_dict)
        print(result_list)
        coordinate = next((r_dict for r_dict in result_list if r_dict.get('text').lower() == input_text.lower()), None)
        return coordinate


METHOD_LIST = ['easy', 'padd']


class OcrTemplate(Template):
    """
    ocr模版
        :param filename: 传入的文字
        :param ocrPlus: 是否启用二值化和高斯模糊
        :param
    """

    ocr_instances = {
        "default": {
            'easy': EasyOcr(),
            'padd': PpOcr()
        }
    }
    lock = threading.Lock()

    def __init__(self, filename, threshold=None, target_pos=TargetPos.MID, record_pos=None, resolution=(), rgb=False,
                 scale_max=800, scale_step=0.005, ocrPlus=False, language=None):

        super().__init__(filename, threshold, target_pos, record_pos, resolution, rgb, scale_max, scale_step)
        self.ocrPlus = ocrPlus
        if language is not None:
            self.language = tuple(language)
            with OcrTemplate.lock:
                if self.language not in OcrTemplate.ocr_instances:
                    OcrTemplate.ocr_instances[self.language] = {
                        'easy': EasyOcr(language=self.language),
                        'padd': PpOcr(language=self.language)
                    }
        else:
            self.language = language

    @property
    def filepath(self):
        return self.filename

    def match_in(self, screen):
        match_result = self._cv_match(screen)
        G.LOGGING.debug("match result: %s", match_result)
        if not match_result:
            return None
        focus_pos = TargetPos().getXY(match_result, self.target_pos)
        return focus_pos

    @logwrap
    def _cv_match(self, screen):

        img = screen

        # 灰度化
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        # 增加对比度和亮度
        # contrast_img = cv2.convertScaleAbs(gray, alpha=1.5, beta=0)
        # 创建一个空的数组，数据类型为'float32'
        img_float = np.float32(gray)
        # 缩放像素值到0-1之间
        img_norm = img_float / 255.0
        # 调整对比度
        contrast_img = np.power(img_norm, 2.3)
        # 将像素值缩放回0-255之间并转换为'uint8'
        contrast_img = np.uint8(contrast_img * 255)

        clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
        clahe_img = clahe.apply(contrast_img)
        # cv2.imshow("Image3", clahe_img)
        # cv2.waitKey(0)

        ret = None
        if self.ocrPlus is True:
            # osu
            _, binary_img = cv2.threshold(contrast_img, 127, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)

            # 应用高斯滤波
            clahe_img = cv2.GaussianBlur(binary_img, (5, 5), 0)
            #
            # cv2.imshow("Image3", binary_img)
            # cv2.waitKey(0)

        for method in METHOD_LIST:
            # get function definition and execute:
            if self.language is not None:
                func = OcrTemplate.ocr_instances[self.language][method]
            else:
                func = OcrTemplate.ocr_instances["default"][method]
            if func is None:
                raise InvalidMatchingMethodError(
                    "Undefined method in OCR_METHOD: '%s'" % method)
            else:
                ret = self._try_match(func, language=self.language, img=clahe_img, input_text=self.filename)
            # 先用easyOCR方法失败则会用下一个
            if ret:
                break
        return ret

    @staticmethod
    def _try_match(func, *args, **kwargs):
        G.LOGGING.debug("try match with %s" % func)
        try:
            instance = func
            ret = instance.ocr_match(img=kwargs['img'], input_text=kwargs['input_text'])
        except aircv.NoModuleError as err:
            G.LOGGING.warning(
                "'Easy/ppd' initialization failed. Alternatively, reinstall easycr or PaddleOCR.")
            return None
        except aircv.BaseError as err:
            G.LOGGING.debug(repr(err))
            return None
        else:
            return ret

# 测试代码
# img_text = cv2.imread(r"G:\pyProject\odin-testautomation\TestAutomation\airtestProject\commons\img\odin\test (2).png")
# rett = OcrTemplate("RTT: 142ms", language=['ch']).match_in(img_text)
# print(rett)



# coding=utf-8
import subprocess
import os
import re
import struct
import logging
from airtestProject.airtest.utils.logger import get_logger
from airtestProject.airtest.utils.nbsp import NonBlockingStreamReader
from airtestProject.airtest.utils.safesocket import SafeSocket
from airtestProject.airtest.utils.compat import SUBPROCESS_FLAG

LOGGING = get_logger(__name__)


class MinicapIOS(object):

    """https://github.com/openstf/ios-minicap"""
    CAPTIMEOUT = None

    def __init__(self, udid=None, port=12345):
        super(MinicapIOS, self).__init__()
        self.udid = udid or list_devices()[0]
        print(repr(self.udid))
        self.port = port
        self.resolution = "320x568"
        self.executable = os.path.join(os.path.dirname(__file__), "ios_minicap")
        self.server_proc = None

    def setup(self):
        cmd = [self.executable, "--udid", self.udid, "--port", str(self.port), "--resolution", self.resolution]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                stdin=subprocess.PIPE, creationflags=SUBPROCESS_FLAG)
        nbsp = NonBlockingStreamReader(proc.stdout, print_output=True, name="minicap_sever")
        while True:
            line = nbsp.readline(timeout=10.0)
            if line is None:
                raise RuntimeError("minicap setup error")
            if b"== Banner ==" in line:
                break
        if proc.poll() is not None:
            logging.warn("Minicap server already started, use old one")
        self.server_proc = proc

    def get_frames(self):
        """
        rotation is alwary right on iOS
        """
        s = SafeSocket()
        s.connect(("localhost", self.port))
        t = s.recv(24)
        # minicap info
        print(struct.unpack("<2B5I2B", t))

        while True:
            # recv header, count frame_size
            if self.CAPTIMEOUT is not None:
                header = s.recv_with_timeout(4, self.CAPTIMEOUT)
            else:
                header = s.recv(4)
            if header is None:
                LOGGING.error("minicap header is None")
                # recv timeout, if not frame updated, maybe screen locked
                yield None
            else:
                frame_size = struct.unpack("<I", header)[0]
                # recv image data
                one_frame = s.recv(frame_size)
                yield one_frame
        s.close()


def list_devices():
    cmd = "system_profiler SPUSBDataType"
    ret = subprocess.check_output(cmd, shell=True).strip() or None
    m = re.findall(r"(?:iPhone|iPad).*?Serial Number: (\w+)", ret, re.DOTALL)
    print(m)
    return m


if __name__ == '__main__':
    m = MinicapIOS()
    m.setup()
    gen = m.get_frames()
    for i in range(100):
        img = next(gen)
        with open("name_%s.png" % i, "wb") as f:
            f.write(img)



# -*- coding: utf-8 -*-
import subprocess
import time
import sys
import random
import shutil
import platform
import warnings
import threading
import wda
from copy import copy
from functools import partial
from airtestProject.airtest.core.error import AirtestError, LocalDeviceError
from airtestProject.airtest.utils.snippet import reg_cleanup, make_file_executable, get_std_encoding, kill_proc
from airtestProject.airtest.utils.logger import get_logger
from airtestProject.airtest.utils.retry import retries
from airtestProject.airtest.utils.compat import SUBPROCESS_FLAG
from airtestProject.airtest.core.ios.constant import DEFAULT_IPROXY_PATH
from airtestProject.airtest.core.ios.relay import ThreadedTCPServer, TCPRelay

LOGGING = get_logger(__name__)


class InstructHelper(object):
    """
    ForwardHelper class
    or help run other Instruction
    """

    def __init__(self, uuid=None):
        self.subprocessHandle = []
        # uuid不是ios手机序列号，而是wda.info['uuid']字段
        self.uuid = uuid
        # 下面_udid这个是真正的手机序列号，检测到是USB设备，才会记录这个字段
        self._udid = None
        self._device = None
        # 记录曾经使用过的端口号和关闭端口用的方法，方便后续释放端口, _port_using_func[port] = kill_func
        self._port_using_func = {}
        reg_cleanup(self.tear_down)

    @staticmethod
    def builtin_iproxy_path():
        # Port forwarding for iOS:
        # 1. Windows/Mac: iproxy.exe/iproxy -u uid port1 port2
        # 2. Ubuntu linux: apt-get install libusbmuxd-tools; iproxy port1 port2
        # 3. Use python (low efficiency): python relay.py -t 5100:5100
        if shutil.which("iproxy"):
            return shutil.which("iproxy")
        system = platform.system()
        iproxy_path = DEFAULT_IPROXY_PATH.get(system)
        if iproxy_path:
            if system == "Darwin":
                make_file_executable(iproxy_path)
            return iproxy_path
        warnings.warn("Please install iproxy for a better experience(Ubuntu Linux): apt-get install libusbmuxd-tools")
        return None

    @property
    def usb_device(self):
        """
        Whether the current iOS uses the local USB interface, if so, return the wda.usbmux.Device object
        当前iOS是否使用了本地USB接口，如果是，返回wda.usbmux.Device对象
        Returns: wda.usbmux.Device or None

        """
        if not self._device:
            # wda无法直接获取iOS的udid，因此先检查usb连接的手机udid列表
            try:
                device_list = wda.usbmux.Usbmux().device_list()
            except ConnectionRefusedError:
                # windows上必须要先启动iTunes才能获取到iOS设备列表
                LOGGING.warning("If you are using iOS device in windows, please check if iTunes is launched")
                return None
            except Exception as e:
                # 其他异常，例如 socket unix:/var/run/usbmuxd unable to connect
                print(e)
                return None
            for dev in device_list:
                udid = dev.get('SerialNumber')
                usb_dev = wda.Client(url=wda.requests_usbmux.DEFAULT_SCHEME + udid)
                # 对比wda.info获取到的uuid是否一致
                try:
                    if usb_dev.info['uuid'] == self.uuid:
                        self._device = wda.usbmux.Device(udid)
                        self._udid = udid
                except:
                    return None
        return self._device

    def tear_down(self):
        # 退出时的清理操作，目前暂时只有清理端口
        using_ports = copy(list(self._port_using_func.keys()))
        for port in using_ports:
            try:
                kill_func = self._port_using_func.pop(port)
                kill_func()
            except:
                continue

    # this function auto gen local port
    @retries(3)
    def setup_proxy(self, device_port):
        """
        Map a port number on an iOS device to a random port number on the machine
        映射iOS设备上某个端口号到本机的随机端口号

        Args:
            device_port: The port number to be mapped on the ios device

        Returns:

        """
        if not self.usb_device:
            raise LocalDeviceError("Currently only supports port forwarding for locally connected iOS devices")
        local_port = random.randint(11111, 20000)
        self.do_proxy(local_port, device_port)
        return local_port, device_port

    def remove_proxy(self, local_port):
        if local_port in self._port_using_func:
            kill_func = self._port_using_func.pop(local_port)
            kill_func()

    def do_proxy(self, port, device_port):
        """
        Start do proxy of ios device and self device
        目前只支持本地USB连接的手机进行端口转发，远程手机暂时不支持
        Returns:
            None

        """
        if not self.usb_device:
            raise LocalDeviceError("Currently only supports port forwarding for locally connected iOS devices")
        proxy_process = self.builtin_iproxy_path() or shutil.which("tidevice")
        if proxy_process:
            cmds = [proxy_process, "-u", self._udid, str(port), str(device_port)]
        else:
            # Port forwarding using python
            self.do_proxy_usbmux(port, device_port)
            return

        # Port forwarding using iproxy
        # e.g. cmds=['/usr/local/bin/iproxy', '-u', '00008020-001270842E88002E', '11565', '5001']
        proc = subprocess.Popen(
            cmds,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=SUBPROCESS_FLAG
        )
        # something like port binding fail
        time.sleep(0.5)

        if proc.poll() is not None:
            stdout, stderr = proc.communicate()
            stdout = stdout.decode(get_std_encoding(sys.stdout))
            stderr = stderr.decode(get_std_encoding(sys.stderr))
            raise AirtestError((stdout, stderr))

        self._port_using_func[port] = partial(kill_proc, proc)

    def do_proxy_usbmux(self, lport, rport):
        """
        Mapping ports of local USB devices using python multithreading
        使用python多线程对本地USB设备的端口进行映射（当前仅使用USB连接一台iOS时才可用）

        Args:
            lport: local port
            rport: remote port

        Returns:

        """
        server = ThreadedTCPServer(("localhost", lport), TCPRelay)
        server.rport = rport
        server.device = self.usb_device
        server.bufsize = 128
        self.server = server
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.daemon = True
        self.server_thread.start()
        self._port_using_func[lport] = self.server.shutdown


if __name__ == '__main__':
    ins = InstructHelper()
    ins.do_proxy_usbmux(5001, 5001)



# coding=utf-8

import time
import threading
import uuid

from airtestProject.poco.sdk.std.transport import Transport
from airtestProject.poco.utils import six
from airtestProject.poco.utils.net.transport.simple_wss import SimpleWebSocketServer, WebSocket

if six.PY3:
    from queue import Queue, Empty
else:
    from Queue import Queue, Empty


class WsSocket(Transport):
    def __init__(self):
        super(WsSocket, self).__init__()
        self.s = None
        self.connections = {}  # websocket client object -> cid
        self.connections_endpoints = {}  # endpoint -> websocket client object
        self.rq = Queue()

    def connect(self, endpoint):
        raise NotImplementedError

    def disconnect(self, endpoint=None):
        raise NotImplementedError

    def bind(self, endpoint):
        if self.s is not None:
            raise RuntimeError("Already bound at {}".format((self.s.serversocket.getsockname())))

        ip, port = endpoint
        if ip in ('', '*', '0'):
            ip = '0.0.0.0'

        class MyWsApp(WebSocket):
            def handleConnected(self2):
                self.connections[self2] = six.text_type(uuid.uuid4())
                self.connections_endpoints[self2.address] = self2
                print('server on accept. {}'.format(self2))

            def handleMessage(self2):
                print('received_message from {}: {}'.format(self2.address, self2.data))
                cid = self.connections.get(self2)
                if cid:
                    self.rq.put((cid, self2.data))

            def handleClose(self2):
                self.connections.pop(self2, None)
                self.connections_endpoints.pop(self2.address, None)
                print('client gone. {}'.format(self2.address))

        self.s = SimpleWebSocketServer(ip, port, MyWsApp)
        t = threading.Thread(target=self.s.serveforever)
        t.daemon = True
        t.start()
        print('server listens on ("{}", {}) transport websocket'.format(ip, port))

    def update(self, timeout=0.001):
        time.sleep(timeout)
        return self.recv()

    def recv(self):
        try:
            return self.rq.get(False)
        except Empty:
            return None, None

    def send(self, cid, data):
        if cid is None:
            # broadcast
            for conn, _ in self.connections.items():
                conn.sendMessage(data)
        else:
            conn = self.get_connection(cid)
            if conn:
                conn.sendMessage(data)

    def get_connection(self, cid_):
        for conn, cid in self.connections.items():
            if cid == cid_:
                return conn
        return None

    def __str__(self):
        return 'Websocket connection(s) at {}'.format(self.connections_endpoints.keys())

    __repr__ = __str__



# _*_ coding:UTF-8 _*_

import os
import json
import time
import inspect
import functools
import traceback
from copy import copy
from .logger import get_logger
from .snippet import reg_cleanup
LOGGING = get_logger(__name__)


class AirtestLogger(object):
    """logger """
    def __init__(self, logfile):
        super(AirtestLogger, self).__init__()
        self.running_stack = []
        self.logfile = None
        self.logfd = None
        self.set_logfile(logfile)
        reg_cleanup(self.handle_stacked_log)

    def set_logfile(self, logfile):
        if logfile:
            self.logfile = os.path.realpath(logfile)
            self.logfd = open(self.logfile, "w")
        else:
            # use G.LOGGER.set_logfile(None) to reset logfile
            self.logfile = None
            if self.logfd:
                self.logfd.close()
                self.logfd = None

    @staticmethod
    def _dumper(obj):
        if hasattr(obj, "to_json"):
            try:
                return obj.to_json()
            except:
                repr(obj)
        try:
            d = copy(obj.__dict__)
            try:
                d["__class__"] = obj.__class__.__name__
            except AttributeError:
                pass
            return d
        except (AttributeError, TypeError):
            # use d = obj.__dict__.copy() to avoid TypeError: can't pickle mappingproxy objects
            # but repr(obj) is simpler in the report
            return repr(obj)

    def log(self, tag, data, depth=None, timestamp=None):
        ''' Not thread safe '''
        # LOGGING.debug("%s: %s" % (tag, data))
        if depth is None:
            depth = len(self.running_stack)
        if self.logfd:
            # 如果timestamp为None，或不是float，就设为默认值time.time()
            try:
                timestamp = float(timestamp)
            except (ValueError, TypeError):
                timestamp = time.time()
            try:
                log_data = json.dumps({'tag': tag, 'depth': depth, 'time': timestamp,
                                       'data': data}, default=self._dumper)
            except UnicodeDecodeError:
                # PY2
                log_data = json.dumps({'tag': tag, 'depth': depth, 'time': timestamp,
                                       'data': data}, default=self._dumper, ensure_ascii=False)
            self.logfd.write(log_data + '\n')
            self.logfd.flush()

    def handle_stacked_log(self):
        # 处理stack中的log
        while self.running_stack:
            # 先取最后一个，记了log之后再pop，避免depth错误
            log_stacked = self.running_stack[-1]
            try:
                self.log("function", log_stacked)
            except Exception as e:
                LOGGING.error("log_stacked error: %s" % e)
                LOGGING.error(traceback.format_exc())
            self.running_stack.pop()


def Logwrap(f, logger):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        """
        The decorator @logwrap can record the function call information in the airtest log and display it in the report.
        装饰器@logwrap，能够在airtest的log中记录函数的调用信息，显示在报告中

        The following parameters can be appended to the function parameter definition for additional effect:
        在函数参数定义中可以附加以下参数，以获得更多效果：

        snapshot: snapshot: If True, a snapshot can be attached to the report. 如果为True，可以附加一张截图到报告中
        depth: the depth order of the current log in the log. 指定log中当前log的深度顺序

        Examples:

            @logwrap
            def func1():
                pass

            @logwrap
            def func1(snapshot=True):
                pass

        Args:
            *args:
            **kwargs:

        Returns:


        """
        from airtestProject.airtest.core.cv import try_log_screen
        # py3 only: def wrapper(*args, depth=None, **kwargs):
        depth = kwargs.pop('depth', None)  # For compatibility with py2
        start = time.time()
        m = inspect.getcallargs(f, *args, **kwargs)
        # The snapshot parameter is popped from the function parameter,
        # so the function cannot use the parameter name snapshot later
        snapshot = m.pop('snapshot', False)
        fndata = {'name': f.__name__, 'call_args': m, 'start_time': start}
        logger.running_stack.append(fndata)
        try:
            res = f(*args, **kwargs)
            if f.__name__ == "try_log_screen":
                data = {"traceback": "疑似资源缺失", "end_time": time.time()}
                fndata.update(data)
        except Exception as e:
            data = {"traceback": traceback.format_exc(), "end_time": time.time()}
            fndata.update(data)
            raise
        else:
            fndata.update({'ret': res, "end_time": time.time()})
            return res
        finally:
            if snapshot is True:
                # If snapshot=True, save an image unless ST.SAVE_IMAGE=False
                try:
                    try_log_screen(depth=len(logger.running_stack) + 1)
                except AttributeError:
                    # if G.DEVICE is None
                    pass
            logger.log('function', fndata, depth=depth)
            try:
                logger.running_stack.pop()
            except IndexError:
                # logger.running_stack is empty
                pass
    return wrapper
